# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for calculating interfaces for CSP"""
import argparse
import sys
from pathlib import Path
from typing import List, Optional

from powertrain_build.interface.service import get_service
from powertrain_build.lib import logger
from powertrain_build.interface import generation_utils

LOGGER = logger.create_logger("CSP service")

PARSER_HELP = "Generate CSP service models"


def configure_parser(parser: argparse.ArgumentParser):
    """Configure parser for CSP service generation"""
    generation_utils.add_base_args(parser)
    parser.add_argument(
        "--client-name",
        help="Name of the context object in CSP. Defaults to project name."
    )
    parser.add_argument(
        "output",
        help="Output directory for service models",
        type=Path
    )
    parser.set_defaults(func=generate_service_cli)


def generate_service_cli(args: argparse.Namespace):
    """CLI function for CSP service generation"""
    app = generation_utils.process_app(args.config)
    client_name = generation_utils.get_client_name(args, app)
    service(args, app, client_name)


def main(argv: Optional[List[str]] = None):
    """ Main function for stand alone execution.
    Mostly useful for testing and generation of dummy hal specifications
    """
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


def service(args, app, client_name):
    """ Generate specifications for pt-scheduler wrappers.

    Args:
        args (Namespace): Arguments from command line
        app (Application): Application to generate specifications for
        client_name (str): Signal client name
    """
    model_internal = get_service(app, client_name, 'internal')
    model_external = get_service(app, client_name, 'external')
    model_observer = get_service(app, client_name, 'observer')
    generation_utils.write_to_file(model_internal, Path(args.output, 'model', 'internal.yaml'), is_yaml=True)
    generation_utils.write_to_file(model_external, Path(args.output, 'model', 'external.yaml'), is_yaml=True)
    generation_utils.write_to_file(model_observer, Path(args.output, 'model', 'observer.yaml'), is_yaml=True)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Performs compatibility upgrade of models to PyBuild using Matlab."""

import argparse
import hashlib
import json
import os
import sys
from pathlib import Path
from re import search
from typing import List, Optional

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
import shutil

import git
from powertrain_build import build, pt_matlab
from powertrain_build.lib import logger, helper_functions

LOGGER = logger.create_logger(__file__)


class PyBuildWrapper(pt_matlab.Matlab):
    """Performs upgrade of Matlab models to PyBuild system."""

    HASH_FILE_NAME = "pybuild_file_hashes.json"
    PARSER_HELP = "Run PyBuild, update and/or generate code for selected models and/or build."

    def __init__(self, args):
        """Constructor, initializes paths for PyBuild upgrader.

        Args:
            args (argument parser): see add_args static method.
        """
        super().__init__(dry_run=args.dry_run, matlab_bin=args.matlab_bin, nojvm=False)

        self.root_path = helper_functions.get_repo_root()
        self.repo = git.Repo(self.root_path)

        self.target_link_settings_hash_file_path = Path("ConfigDocuments", "target_link_settings_file_hashes.json")
        self.target_link_settings_folders = [
            Path("ConfigDocuments", "targetlinkSettings"),
            Path("ConfigDocuments", "TL4_3_settings"),
        ]

        self.update = args.update
        self.codegen = args.codegen

        self.build = args.build
        self.build_specific = getattr(args, "project_config", None) is not None
        self.project_config = self._set_project_configuration(args)
        self.generate_system_info = getattr(args, "generate_system_info", False)
        self.core_dummy = getattr(args, "core_dummy", True)
        self.rte_dummy = getattr(args, "rte_dummy", False)
        self.debug = getattr(args, "debug", True)
        self.no_abort = getattr(args, "no_abort", True)
        self.no_nvm_a2l = getattr(args, "no_nvm_a2l", False)
        self.complete_a2l = getattr(args, "complete_a2l", False)
        self.silver_a2l = getattr(args, "silver_a2l", False)
        self.generate_rte_checkpoint_calls = getattr(args, "generate_rte_checkpoint_calls", False)
        self.interface = args.interface
        self.matlab_include = args.include
        # Always default to conversion table in pytools/config/conversion_table.json.
        # Unless otherwise specified.
        conv_tab_path = getattr(args, "generate_custom_conv_tab", None)
        default_conv_tab_path = os.path.join("ConfigDocuments", "pytools_settings", "conversion_table.json")
        self.conv_tab = conv_tab_path if conv_tab_path is not None else default_conv_tab_path

        self.should_run, self.models = self._evaluate_run_and_models(args)

    @staticmethod
    def _set_project_configuration(args):
        """Evaluate path to project configuration file.

        Args:
            args (argument parser): see add_args static method.
        Returns:
            project_config: Path to project configuration file.
        """
        if getattr(args, "project_config", None) is not None:
            project_config = args.project_config.replace("/", os.sep)
        elif args.build is not None:
            if args.build.lower() == "custom":
                # TODO: Change other scripts to accept custom config
                #       Change this to point to a config file instead
                project_config = os.path.join(os.environ.get("PROJECT_DIR"), "ProjectCfg.json")
            else:
                project_config = os.path.join("Projects", args.build, "ProjectCfg.json")
        else:
            project_config = None

        return project_config

    @staticmethod
    def convert_path_sep(paths):
        """Matlab requires forward-spaces for model paths. Convert backslashes.

        Args:
            paths (list): Model paths to fix path separators for.
        Returns:
            (list): Models paths separated with forward slashes.
        """
        return [path.replace("\\", "/") for path in paths]

    @staticmethod
    def get_matlab_scripts_commit_sha():
        """Get current commit sha of matlab-scripts submodule, if available.

        Returns:
            matlab_scripts_commit_sha (str): Commit sha of matlab-scripts submodule.
                None, if not available.
        """
        repo = git.Repo()
        try:
            matlab_scripts = repo.submodule("matlab-scripts")
            matlab_scripts_commit_sha = matlab_scripts.hexsha
        except ValueError:
            LOGGER.info("Submodule matlab-scripts not available, " "skipping adding its commit sha to hash file.")
            matlab_scripts_commit_sha = None
        return matlab_scripts_commit_sha

    @staticmethod
    def read_bytes(file_path):
        """Read file contents in byte mode.

        Args:
            file_path (Path): Path to file.
        Returns:
            file_bytes (bytes): Content of file read in binary mode.
        """
        if file_path.suffix in [".slx", ".mdl", ".mexw64"]:
            # Already treated as binary file
            file_bytes = file_path.read_bytes()
        else:
            # Hack to make this script os independent
            with file_path.open(encoding="iso-8859-1") as file_handle:
                content = file_handle.read()
                file_bytes = content.encode("iso-8859-1")
        return file_bytes

    @staticmethod
    def get_files_to_hash(model_folder):
        """Get files to generate hashes for (PyBuild specific files).

        Args:
            model_folder (Path): Path to model folder.
        Returns:
            files_to_hash (list): List of, pybuild specific, files to generate hashes for.
        """
        valid_file_endings = [
            # source files
            ".a2l",
            ".c",
            ".h",
            ".mexw64",
            ".tlc",
            # config files
            ".json",
            # model files
            ".mdl",
            ".m",
        ]
        valid_folders = [
            model_folder,
            Path(model_folder, "matlab_src"),
            Path(model_folder, "pybuild_cfg"),
            Path(model_folder, "pybuild_src"),
        ]
        hash_file_path = Path(model_folder, PyBuildWrapper.HASH_FILE_NAME)

        model_files = list(model_folder.rglob("*.*"))
        if hash_file_path.exists():
            model_files.remove(hash_file_path)
        files_to_hash = [f for f in model_files if f.parent in valid_folders and f.suffix in valid_file_endings]
        return files_to_hash

    @staticmethod
    def get_shared_function_files():
        """Get shared function files, files generated by VcSharedFunctions.mdl.

        Returns:
            shared_function_files (list): List of files generated by VcSharedFunctions.mdl.
        """
        shared_function_files = []
        common_source_folder = Path("Models/Common/pybuild_src")
        for source_file in common_source_folder.glob("*.*"):
            with source_file.open() as file_handle:
                content = file_handle.read()
            if search(r"Simulink model\s+: VcSharedFunctions", content) is not None:
                shared_function_files.append(source_file)
        return shared_function_files

    @staticmethod
    def get_file_hashes(model_paths, write_to_file=True):
        """Calculate SHA256 file hashes for files in given model folders.

        Args:
            model_paths (list): List of model paths.
            write_to_file (bool): True/False whether calculated hashes should be written to file.
        Returns:
            model_to_files_hash_dict (dict): Dict mapping model name to files and its hashes.
        """
        model_to_files_hash_dict = {}
        model_folders = [(Path(m).parent, Path(m).stem) for m in model_paths]
        for model_folder, model_name in model_folders:
            file_to_hash_dict = {}
            files_to_hash = PyBuildWrapper.get_files_to_hash(model_folder)
            if "VcSharedFunctions" in model_folder.as_posix():
                shared_function_files = PyBuildWrapper.get_shared_function_files()
                files_to_hash.extend(shared_function_files)
            for file_to_hash in files_to_hash:
                file_bytes = PyBuildWrapper.read_bytes(file_to_hash)
                file_to_hash_dict[file_to_hash.name] = hashlib.sha256(file_bytes).hexdigest()
            commit_sha = PyBuildWrapper.get_matlab_scripts_commit_sha()
            if commit_sha is not None:
                file_to_hash_dict["matlab-scripts"] = commit_sha
            if write_to_file:
                with Path(model_folder, PyBuildWrapper.HASH_FILE_NAME).open("w", encoding="utf-8") as file_handle:
                    json.dump(file_to_hash_dict, file_handle, indent=4)
            model_to_files_hash_dict[model_name] = file_to_hash_dict
        return model_to_files_hash_dict

    def _evaluate_run_and_models(self, args):
        """Evaluate if PyBuild Matlab related parts should run.

        Additionally, it sets which models to update and/or generate code for, based on arguments.
        NOTE: model_list=None indicates all models when run is True. Deprecated, add force flag?

        Args:
            args (argument parser): see add_args static method.
        Returns:
            should_run: True/False whether PyBuild should run or not.
            model_list: List of models to update/generate code for.
        """
        run_powertrain_build = args.build is not None or getattr(args, "project_config", None) is not None
        should_run = args.update or args.codegen or run_powertrain_build
        model_list = args.models if args.models else self.get_changed_models()
        LOGGER.info("Affected models: %s", ", ".join(model_list))
        if not model_list:
            # PyBuild should not run if there were no model changes
            should_run = run_powertrain_build
            model_list = None

        return should_run, model_list

    def regenerate_target_link_settings_file_hashes(self):
        """Regenerate the file mapping TargetLink settings files to their hashes."""
        target_link_settings_file_hashes = self.calculate_target_link_settings_file_hashes()
        with self.target_link_settings_hash_file_path.open("w", encoding="utf-8") as file_handle:
            json.dump(target_link_settings_file_hashes, file_handle, indent=4)

    def calculate_target_link_settings_file_hashes(self):
        """Calculate SHA256 file hashes for files in TargetLink settings folders.

        Returns:
            file_to_hash_dict (dict): Dict mapping settings files and their hashes.
        """
        file_to_hash_dict = {}
        for settings_folder in self.target_link_settings_folders:
            for settings_file in settings_folder.rglob("*.*"):
                file_bytes = self.read_bytes(settings_file)
                file_to_hash_dict[settings_file.name] = hashlib.sha256(file_bytes).hexdigest()
        return file_to_hash_dict

    def verify_target_link_settings(self):
        """Verify current TargetLink settings, comparing against commited file hashes file.

        Returns:
            (bool): True/False, depending on if the TargetLink settings have changed.
        """
        if not self.target_link_settings_hash_file_path.exists():
            message = (
                "Could not read TargetLink settings file hashes file: "
                f"{self.target_link_settings_hash_file_path.as_posix()}.\n"
                "If your repo runs the jobb PyBuildDiff in hash mode, make sure to generate one."
                "Ignoring settings verification."
            )
            LOGGER.warning(message)
            return True

        with self.target_link_settings_hash_file_path.open(encoding="utf-8") as file_handle:
            current_file_hashes_dict = json.load(file_handle)

        new_file_hashes_dict = self.calculate_target_link_settings_file_hashes()

        return new_file_hashes_dict == current_file_hashes_dict

    def get_changed_models(self):
        """Get changed models in current commit."""
        changed_files_tmp = self.repo.git.diff("--diff-filter=d", "--name-only", "HEAD~1")
        changed_files = changed_files_tmp.splitlines()
        changed_models = [m for m in changed_files if m.endswith(".mdl") or m.endswith(".slx")]
        return changed_models

    def check_generate_shared_functions(self, matlab_command_list):
        """Check if shared function files should be generated.

        Args:
            matlab_command_list ([str]): list of matlab commands.
        """
        if [model for model in self.models if model.endswith("VcSharedFunctions.mdl")]:
            matlab_command_list.append(pt_matlab.cmd_callfunc("generateSharedFunctions", True))

    def build_automation(self, mode):
        """Run Matlab with a specific task and specific models.

        Args:
            mode (str): Matlab run mode (update, codegen).
        Returns:
            exit_code (int): Exit code from Matlab.
        """
        # Will be used in submodule matlab-scripts (if up to date), CodeGen/updateCodeSwConfig.m
        calling_python = sys.version_info
        calling_python_string = f"py -{calling_python.major}.{calling_python.minor}"
        os.environ.setdefault("CALLING_PYTHON", calling_python_string)

        # Specify a new script and log file name for each run
        script_name = f"powertrain_build_matlab_{mode}.m"
        self.log = f"powertrain_build_matlab_{mode}.log"

        # Set up matlab path
        matlab_scripts_path = os.path.join("powertrain_build_matlab_scripts")

        # Copy matlab-scripts to the project root
        matlab_script_folder = files("powertrain_build.matlab_scripts")
        shutil.rmtree(matlab_scripts_path, ignore_errors=True)
        shutil.copytree(matlab_script_folder, matlab_scripts_path)

        cmds = []
        cmds.append(f"cd '{self.root_path}'")
        cmds.append(pt_matlab.cmd_path(matlab_scripts_path, True))
        cmds.append(pt_matlab.cmd_path(self.matlab_include, True))

        if mode == "codegen":
            cmds.append(pt_matlab.cmd_callfunc("loadLibraries", self.matlab_include))

        # Generate the command for calling the main build script:
        args = [mode, True]
        if self.models:
            self.check_generate_shared_functions(cmds)
            args.append(self.convert_path_sep(self.models))

        cmds.append(pt_matlab.cmd_callfunc("BuildAutomationPyBuild", *args))
        pt_matlab.write_m_script(script_name, pt_matlab.cmds_join(cmds), wrap_cmd=False)

        # Reset the Matlab watcher before running a new m-script
        self.matlab_watcher.reset_errors()

        self.run_m_script(script_name, wrap_cmd=False, attempts=2)

        if not self.matlab_watcher.task_success:
            LOGGER.error("PyBuild %s error!", mode)
            return 1

        if not self.verify_target_link_settings():
            LOGGER.error("Your TargetLink settings differ from the ones set by current branch.")
            return 1

        self.get_file_hashes(self.models)
        return 0

    def run_powertrain_build(self):
        """Execute powertrain-build.

        Returns:
            exit_code: Exit code from powertrain-build build step.
        """
        try:
            exit_code = build.build(
                self.project_config,
                interface=self.interface,
                core_dummy=self.core_dummy,
                rte_dummy=self.rte_dummy,
                no_abort=self.no_abort,
                no_nvm_a2l=self.no_nvm_a2l,
                debug=self.debug,
                generate_system_info=self.generate_system_info,
                generate_custom_conversion_table=self.conv_tab,
                complete_a2l=self.complete_a2l,
                silver_a2l=self.silver_a2l,
                generate_rte_checkpoint_calls=self.generate_rte_checkpoint_calls,
            )
        except (FileNotFoundError, PermissionError) as ex:
            LOGGER.error(ex)
            exit_code = 1
        return exit_code

    def run(self):
        """Run PyBuild, update and/or generate code for selected models and/or build.

        Returns:
            exit_code: Exit code from Matlab and build step.
        """
        if not self.should_run:
            return 0

        LOGGER.info("Preparing workspace for PyBuild!")

        exit_code = 0
        if self.update:
            LOGGER.info("Running PyBuild update!")
            exit_code |= self.build_automation(mode="update")

        if self.codegen:
            LOGGER.info("Running PyBuild generate code!")

            exit_code |= self.build_automation(mode="codegen")

        if self.build:
            LOGGER.info("Running PyBuild.build for %s!", self.build)
            exit_code |= self.run_powertrain_build()

        if self.build_specific:
            LOGGER.info("Running PyBuild specific build for %s!", self.project_config)
            exit_code |= self.run_powertrain_build()

        return exit_code

    @staticmethod
    def add_args(parser):
        """Add expected arguments to the supplied parser.

        Args:
            parser (ArgumentParser): parser to add arguments to.
        """
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Dry run: No changes, simulation only (default: False).",
        )
        parser.add_argument("--repo-root", help="Path to repository where work should be done")
        parser.add_argument(
            "--interface",
            action="store_true",
            help="Create interface consistency report (default: False)" "NOTE: This requires the --build flag.",
        )
        parser.add_argument(
            "--models",
            nargs="+",
            default=None,
            help="List of model files (full path, "
            "e.g. Models/<SSP>/<MODEL>/<MDL-FILE>) to upgrade, separated with "
            "spaces. Takes precedence over Git HEAD.",
        )
        parser.add_argument(
            "--update",
            action="store_true",
            help="Run PyBuild update on models (default: False).",
        )
        parser.add_argument(
            "--codegen",
            action="store_true",
            help="Run PyBuild code generation on models (default: False). "
            "NOTE: This requires either the --update flag or already updated models.",
        )
        parser.add_argument(
            "--build",
            default=None,
            help="Run PyBuild for project with standard settings, VCC SPM SW release " "(default: None).",
        )
        parser.add_argument(
            "--regenerate-tl-settings-hashes",
            action="store_true",
            help="Regenerate the file mapping TargetLink settings files to their hashes.",
        )
        parser.add_argument(
            "--include",
            default="matlab-scripts",
            help="Path to folder containing Matlab scripts and simulink libraries to include.",
        )
        parser.set_defaults(func=run_wrapper)

        subparsers = parser.add_subparsers(help="PyBuild specific build.")
        build_specific_parser = subparsers.add_parser(
            "build-specific", help="Run PyBuild for project with specific settings."
        )
        build.add_args(build_specific_parser)

        # Matlab arguments added by parent:
        pt_matlab.Matlab.add_args(parser)


def run_wrapper(args: argparse.Namespace) -> int:
    """Run PyBuildWrapper."""
    if args.build is not None and getattr(args, "project_config", None) is not None:
        LOGGER.error("Cannot run both PyBuild quick build (--build <PROJECT>) " "and specific build (build-specific).")
        return 1

    wrapper = PyBuildWrapper(args)

    if args.regenerate_tl_settings_hashes:
        wrapper.regenerate_target_link_settings_file_hashes()
        return 0

    return wrapper.run()


def main(argv: Optional[List[str]] = None) -> int:
    """Run PyBuildWrapper"""
    parser = argparse.ArgumentParser("PyBuild Wrapper")
    PyBuildWrapper.add_args(parser)
    args = parser.parse_args(argv)
    return run_wrapper(args)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for html report generation."""
from string import Template


class HtmlReport:
    """Generate template html report. Extend this class to add content.

    TODO: Refactor common parts from derived report classes to this class.
    """

    __html_head = """<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<style>
    body {
        counter-reset: h2;
        font-family:"Segoe UI","Segoe",Tahoma,Helvetica,Arial,sans-serif;
    }
    h1 {text-align: center}
    h2 {counter-reset: h3}
    h3 {counter-reset: h4}
    h4 {counter-reset: h5}
    h5 {counter-reset: h6}
    h2:before {counter-increment: h2; content: counter(h2) ". "}
    h3:before {counter-increment: h3; content: counter(h2) "." counter(h3) ". "}
    h4:before {counter-increment: h4; content: counter(h2) "." counter(h3) "." counter(h4) ". "}
    h5:before {counter-increment: h5; content: counter(h2) "." counter(h3) "." counter(h4) "." counter(h5) ". "}
    h6:before {
      counter-increment: h6; content: counter(h2) "." counter(h3) "." counter(h4) "." counter(h5) "." counter(h6) ". "
    }
    h2.nocount:before, h3.nocount:before, h4.nocount:before, h5.nocount:before, h6.nocount:before {
      content: ""; counter-increment: none
    }
    p {
        padding-left:5mm;
    }
    table {
        width:95%;
        margin-left:5mm
    }
    table, th, td {
        border: 1px solid black;
        border-collapse: collapse;
    }
    th, td {
        padding: 5px;
        text-align: left;
    }
    table tr:nth-child(even) {
        background-color: #eee;
    }
    table tr:nth-child(odd) {
        background-color:#fff;
    }
    table th {
        background-color: black;
        color: white;
    }
    ol {
        list-style-type: none;
        counter-reset: item;
        margin: 0;
        padding-left:5mm;
    }
    ol > li {
        display: table;
        counter-increment: item;
        margin-bottom: 0.6em;
    }
    ol > li:before {
        content: counters(item, ".") ". ";
        display: table-cell;
        padding-right: 0.6em;
    }
    li ol > li {
        margin: 0;
    }
    li ol > li:before {
        content: counters(item, ".") " ";
    }
</style>
<title>$title</title>
</head>
<body>
<h1>$title</h1>
"""

    __html_end = """</body>
</html>
"""

    def __init__(self, title):
        """Initialize report title.

        Args:
            title (str): The report title
        """
        super().__init__()
        self._title = title

    def _gen_header(self):
        """Generate html header."""
        return Template(self.__html_head).substitute(title=self._title)

    @staticmethod
    def gen_contents():
        """Generate report contents.

        Override this method to add report contents.
        Contents should start with a <h2> header.
        """
        return '<h2>Template report</h2>'

    def _gen_end(self):
        """Generate the end of the body and html document."""
        return self.__html_end

    def generate_report_string(self):
        """Generate a html report as string."""
        html = []
        html += self._gen_header()
        html += self.gen_contents()
        html += self._gen_end()
        return ''.join(html)

    def generate_report_file(self, filename):
        """Generate a html report and save to file."""
        with open(filename, 'w', encoding="utf-8") as fhndl:
            fhndl.write(self.generate_report_string())



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for reading unit configuration files."""
import json
import os
import time
from pprint import pformat

from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.versioncheck import Version


class CodeGenerators:
    """Enum for code generators."""
    target_link = 'target_link'
    embedded_coder = 'embedded_coder'


class UnitConfigs(ProblemLogger):
    """A class for accessing the project’s unit definitions (see :doc:`unit_config`).

    Provides methods for retrieving the all definitions of a unit and all existing units.
    """

    CONFIG_SKIP_LIST = ['VcDebugSafe', 'VcDebug', 'VcDebugOutputSafe', 'VcDebugOutput']

    def __init__(self, build_prj_config, feature_config):
        """Class Initialization.

        Args:
            build_prj_config (BuildProjConfig): A class instance which holds
                                                the information of where to find units configs to parse
            feature_config (FeatureConfigs): Class instance project feature definitions

        """
        super().__init__()
        if not isinstance(build_prj_config, BuildProjConfig):
            raise TypeError('build_prj_config argument is not an'
                            ' instance of BuildProjConfig')
        self._build_prj_config = build_prj_config
        self._feature_cfg = feature_config
        self._raw_per_unit_configs = {}
        self._per_unit_configs = {}
        self._per_type_unit_configs = {}
        self._if_define_dict = {}
        self._missing_configs = set()
        self._empty_config_def = set()
        self._parse_all_unit_configs()
        self._per_type_unit_signals()
        self.code_generators = self._get_code_generators()
        self.base_types_headers = self._get_base_types_headers()

        # write the summary of error (to avoid repeating error messages)
        for unit in self._missing_configs:
            self.critical('%s is missing config files', unit)
        for var, unit in self._empty_config_def:
            self.warning('%s in unit %s, has empty config_def!'
                         'probably goto-block, missing a corresponing from-block.',
                         var, unit)

    def __repr__(self):
        """Get string representation of object."""
        return pformat(self._per_type_unit_configs)

    def _parse_all_unit_configs(self):
        """Parse all unit config files."""
        start_time = time.time()
        self.info('  Start loading unit_cfg json files')
        cfg_dirs = self._build_prj_config.get_unit_cfg_dirs()
        for unit, cfg_dir in cfg_dirs.items():
            self._parse_unit_config(unit, cfg_dir)
        self.info('  Finished loading unit_cfg json files (in %4.2f s)', time.time() - start_time)

    def _parse_unit_config(self, unit, cfg_dir):
        """Parse one unit config file."""
        file_ = os.path.join(cfg_dir, f'config_{unit}.json')
        with open(file_, 'r', encoding="utf-8") as fhndl:
            self.debug('Loading json file %s', unit)
            try:
                tmp_ucfg = json.load(fhndl)
                if not Version.is_compatible(tmp_ucfg.get('version', '0.0.0')):
                    raise ValueError(f'Incompatible config file version for unit {unit}.')
                if unit in self._raw_per_unit_configs:
                    self.critical("Conflicting Unit name %s: Units need to have unique names", unit)
                self._raw_per_unit_configs[unit] = tmp_ucfg
            except json.JSONDecodeError as ex:
                self.critical('Error reading config file %s: %s', file_, ex)
                return
        for include_unit in tmp_ucfg.get('includes', []):
            self.debug('%s includes %s in %s', unit, include_unit, cfg_dir)
            self._parse_unit_config(include_unit, cfg_dir)

    def _filter_io_nvm_feat(self):
        """Remove all parameters not defined in the prj_config.

        Parameters can be removed via not active feature in the unit, or
        the entire unit is not included in the project.

        Args:
            config (str): the name of the configuration

        the format of the data-dict::

            {'UnitName': {'class': 'CVC_EXT',
                          'configs': [['Vc_D_CodegenHev '
                                      '== 2',
                                      'Vc_D_CodegenHev '
                                      '> 0']],
                          'description': 'HV battery cooling request',
                          'handle': 'VcPpmPsm/VcPpmPsm/Subsystem/'
                                    'VcPpmPsm/yVcBec_B_ChillerCoolReq',
                          'lsb': 1,
                          'max': 3,
                          'min': 0,
                          'name': 'sVcBec_D_HvBattCoolgReq',
                          'offset': 0,
                          'type': 'UInt8',
                          'unit': '-'}
            }

        """
        res = {}
        self.debug('_filter_io_nvm_feat: Feature Cfg')
        for unit in self._build_prj_config.get_included_units():
            self._filter_io_nvm_feat_unit(unit, res)
        if not res:
            self.warning('No units configured for project')
        return res

    def _filter_core_config(self, u_def_data):
        """Handle core configs."""
        f_core = {}
        for core_type, core_data in u_def_data.items():
            f_core[core_type] = {}
            for key, value in core_data.items():
                if key != 'IllegalBlk':
                    # Matlab sets core:{type:{name:{API_blk:[{path, config}]}}}
                    # config.py - core:{type:{name:{configs}}}
                    configs = value.get('configs', [cfg for blk in value['API_blk'] for cfg in blk['config']])
                    if self._feature_cfg.check_if_active_in_config(configs):
                        f_core[core_type][key] = value
        return f_core

    def _filter_io_nvm_feat_unit(self, unit, res):
        """Handle one unit config with respect to the filtering in :_filter_io_nvm_feat:."""
        try:
            u_data = self._raw_per_unit_configs[unit]
        except KeyError:
            # Some units in the raster should not have config files
            if unit not in self.CONFIG_SKIP_LIST:
                self.debug('_filter_io_nvm_feat_unit: cfg missing: %s', unit)
                self._missing_configs.add(unit)
            return
        for u_def_type, u_def_data in u_data.items():
            res.setdefault(unit, {}).setdefault(u_def_type, {})
            if u_def_type == 'dids':
                f_dids = {k: v for k, v in u_def_data.items()
                          if self._feature_cfg.check_if_active_in_config(v['configs'])}
                res[unit][u_def_type] = f_dids
            elif u_def_type == 'core':
                res[unit][u_def_type] = self._filter_core_config(u_def_data)
            elif u_def_type == 'pre_procs':
                # the pre_proc key does not have configuration attributes
                res[unit]['pre_procs'] = u_def_data
            elif u_def_type == 'integrity_level':
                res[unit]['integrity_level'] = u_def_data
            elif u_def_type == 'code_generator':
                res[unit]['code_generator'] = u_def_data
            elif u_def_type == 'version':
                res[unit]['version'] = u_def_data
            elif u_def_type == 'csp':
                csp_data = {}
                if 'methods' in u_def_data:
                    csp_data = {'methods': {}}
                    for method_name, method_data in u_def_data['methods'].items():
                        if self._feature_cfg.check_if_active_in_config(method_data['configs']):
                            csp_data['methods'][method_name] = method_data
                res[unit][u_def_type] = csp_data
            elif u_def_type == 'includes':
                # List of configs for handwritten code
                for included_unit in u_def_data:
                    self.debug('%s includes %s', unit, included_unit)
                    self._filter_io_nvm_feat_unit(included_unit, res)
            else:
                for var, var_pars in u_def_data.items():
                    # TODO: remove this code when the bug in the matlab code is removed.
                    if var_pars['configs'] == []:
                        self.debug('Adding %s', unit)
                        self._empty_config_def.add((var, unit))
                    if self._feature_cfg.check_if_active_in_config(var_pars['configs']):
                        res[unit][u_def_type].setdefault(var, {}).update(var_pars)

    @staticmethod
    def _update_io_nvm(dict_, unit, data_type, variables):
        """Change the struct for in out and nvm variables.

        The resulting new struct is stored in the dict dict_
        """
        for var, var_pars in variables.items():
            dict_.setdefault(data_type, {}).setdefault(var, {}).setdefault(unit, var_pars)

    def _update_dids(self, unit, key, data, feat_cfg=None):
        """Change the struct for in out and nvm variables."""
        # TODO: Add functionality

    @staticmethod
    def _update_core(dict_, unit, data_type, core_ids):
        """Change the struct for in core parameters."""
        for _, core_data in core_ids.items():
            for var, var_pars in core_data.items():
                dict_.setdefault(data_type, {}).setdefault(var, {}).setdefault(unit, var_pars)

    def _update_pre_procs(self, unit, key, data, feat_cfg=None):
        """Change the struct for in pre_processor parameters."""
        # TODO: Add functionality

    def _per_type_unit_signals(self):
        """Change the structure of the data to aggregate all unit configs.

        Returns:
            dict: a structure per config type instead of per unit

        """
        # loop over all projects and store the active items in each configuration
        self._per_unit_configs = self._filter_io_nvm_feat()
        dict_ = self._per_type_unit_configs = {}
        for unit, udata in self._per_unit_configs.items():
            for data_type, variables in udata.items():
                if data_type in ['core']:
                    self._update_core(dict_, unit, data_type, variables)
                elif data_type in ['dids']:
                    self._update_dids(dict_, unit, data_type, variables)
                elif data_type in ['pre_procs']:
                    self._update_pre_procs(dict_, unit, data_type, variables)
                elif data_type in ['outports', 'inports', 'dids', 'nvm', 'local_vars', 'calib_consts', 'csp']:
                    self._update_io_nvm(dict_, unit, data_type, variables)
                else:
                    dict_.setdefault(data_type, {}).setdefault(unit, udata)

    def check_if_in_unit_cfg(self, unit, symbol):
        """Check if the symbol is defined in the unit config file."""
        for data in self._raw_per_unit_configs[unit].values():
            if isinstance(data, dict):
                if symbol in data:
                    return True
        return False

    def get_per_cfg_unit_cfg(self):
        """Get all io-signals and core-ids for all units.

        Get all io-signals and core-ids for all units, where all inports, outport, etc,
        are aggregated from all unit definition files.

        Returns:
            dict: a dict with the below format::

            {
                'inports/outports/nvm/core': {
                    'VARIABLE_NAME': {
                        'UNIT_NAME': {
                            'class': 'CVC_EXT',
                            'configs': [['all']],
                            'description': 'Power Pulse ',
                            'handle': 'VcPemAlc/VcPemAlc/Subsystem/VcPemAlc/yVcVmcPmm_B_SsActive9',
                            'lsb': 1,
                            'max': 800,
                            'min': 0,
                            'name': 'sVcAesPp_Pw_PwrPls',
                            'offset': 0,
                            'type': 'UInt16',
                            'unit': 'W'
                            }
                        }
                    }
                }
            }

        The top level keys are 'inports', 'outports', 'nvm' and 'core'

        """
        return self._per_type_unit_configs

    def check_if_in_per_cfg_unit_cfg(self, cfg, symbol):
        """Check if the symbol is defined in the aggregated unit config files."""
        return (
                cfg in self._per_type_unit_configs and
                symbol in self._per_type_unit_configs[cfg])

    def get_per_unit_cfg(self):
        """Get io-signals for all units, per unit, for a given project.

        If 'all' is given as a project, all signals, regardless of configuration,
        is returned.

        Returns:
            dict: a dict with the below format::

            {'NAME_OF_UNIT': {
                'core': {'Events': {},
                         'FIDs': {},
                         'IUMPR': {},
                         'Ranking': {},
                         'TstId': {}},
                'dids': {},
                'inports': {'VARIABLE_NAME': {'class': 'CVC_DISP',
                                                        'configs': [['all']],
                                                        'description': 'Torque '
                                                                        'arbitraion '
                                                                        'state',
                                                        'handle': 'VcPemAlc/VcPemAlc/...',
                                                        'lsb': 1,
                                                        'max': 9,
                                                        'min': 0,
                                                        'name': 'rVcPemAlc_D_AuxLoadEvent',
                                                        'offset': 0,
                                                        'type': 'UInt8',
                                                        'unit': '-'}
                            },
                'outports' : {},
                'nvm' : {}
                }
            }

        """
        return self._per_unit_configs

    def get_per_unit_cfg_total(self):
        """Get total io-signals configuration for all units, per unit, for a given project.

        Does not remove signals disabled by code switches.

        Returns:
            dict: a dict with the below format::

            {'NAME_OF_UNIT': {
                'core': {'Events': {},
                         'FIDs': {},
                         'IUMPR': {},
                         'Ranking': {},
                         'TstId': {}},
                'dids': {},
                'inports': {'VARIABLE_NAME': {'class': 'CVC_DISP',
                                                        'configs': [['all']],
                                                        'description': 'Torque '
                                                                        'arbitraion '
                                                                        'state',
                                                        'handle': 'VcPemAlc/VcPemAlc/...',
                                                        'lsb': 1,
                                                        'max': 9,
                                                        'min': 0,
                                                        'name': 'rVcPemAlc_D_AuxLoadEvent',
                                                        'offset': 0,
                                                        'type': 'UInt8',
                                                        'unit': '-'}
                            },
                'outports' : {},
                'nvm' : {}
                }
            }

        """
        res = {}
        for unit in self._build_prj_config.get_included_units():
            try:
                res[unit] = self._raw_per_unit_configs[unit]
            except KeyError:
                continue
        return res

    def get_unit_config(self, unit):
        """Get config for a unit.

        Arguments:
        unit (str): Unit to get config for
        """
        return self._raw_per_unit_configs[unit]

    @staticmethod
    def get_base_name(unit):
        """Get base name of unit."""
        return unit.partition('__')[0]

    def get_unit_code_generator(self, unit):
        """Get code generator for a given a unit (model).

        Args:
            unit (str): Current unit/model name.
        Returns:
            code_generator (str): Code generator used for given model.
        """
        per_unit_cfg = self.get_per_unit_cfg()
        if unit in per_unit_cfg and 'code_generator' in per_unit_cfg[unit]:
            code_generator = per_unit_cfg[unit]['code_generator']
        else:
            # Default to target_link
            code_generator = CodeGenerators.target_link
        return code_generator

    def _get_code_generators(self):
        per_unit_cfg = self.get_per_unit_cfg()
        code_generators = set()
        for _, config in per_unit_cfg.items():
            if 'code_generator' in config:
                code_generators.add(config['code_generator'])
            else:
                # Default to target_link
                code_generators.add(CodeGenerators.target_link)
        return code_generators

    def _get_base_types_headers(self):
        general_includes = ''
        if CodeGenerators.embedded_coder in self.code_generators:
            general_includes += '#include "rtwtypes.h"\n'
        if CodeGenerators.target_link in self.code_generators:
            general_includes += '#include "tl_basetypes.h"\n'
        return general_includes



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Wrapper for win32api to simplify it's use.

See pywin32 documentation or windows api documentation for more information.
"""

from powertrain_build.lib.helper_functions import merge_dicts

STILL_ACTIVE = 259

try:
    import win32job
    import win32process

    class PtWin32Process:
        """Alternative to python subprocess.

        This adds functionality to add processes to win32 job containers
        to force processes to terminate after it's parent process dies.
        """

        def __init__(self, kill_on_job_close=True):
            """Set up default values.

            arguments:
            kill_on_job_close (boolean): If True the process will be terminated when all job
                                         handles are closed.
            """
            self.process_handle = None
            self.thread_handle = None
            self.job_handle = None
            self.process_id = None
            self.thread_id = None
            self.process_attributes = None
            self.thread_attributes = None
            self.b_inherit_handles = True
            self.dw_creation_flags = 0
            self.new_environment = None
            self.current_directory = None
            self.job_limit_flag_list = []
            self.startup_info = win32process.STARTUPINFO()
            self.limit_info_dict = {"BasicLimitInformation": {"LimitFlags": 0}}

            if kill_on_job_close:
                self.add_limit_flags(win32job.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE)

        def run(self, app_name=None, command=None):
            """Run process.

            Can be run with either a specified application, a command line string or both.

            Keyword arguments:
            app_name (string): path to application to execute.
            command (string): command line string to use in execution.
            """
            self.process_handle, self.thread_handle, self.process_id, self.thread_id = win32process.CreateProcess(
                app_name,
                command,
                self.process_attributes,
                self.thread_attributes,
                self.b_inherit_handles,
                self.dw_creation_flags,
                self.new_environment,
                self.current_directory,
                self.startup_info,
            )

            self.job_handle = win32job.CreateJobObject(None, "")
            limit_information = self.get_extended_limit_information()
            limit_information = merge_dicts(self.limit_info_dict, limit_information, merge_recursively=True)
            win32job.SetInformationJobObject(
                self.job_handle, win32job.JobObjectExtendedLimitInformation, limit_information
            )
            win32job.AssignProcessToJobObject(self.job_handle, self.process_handle)

        def get_extended_limit_information(self):
            """Get extended limit information."""
            return win32job.QueryInformationJobObject(self.job_handle, win32job.JobObjectExtendedLimitInformation)

        def set_current_directory(self, current_directory):
            """Set process working directory."""
            self.current_directory = current_directory

        def set_environment(self, environment):
            """Set process environment variables.

            If not set the process will inherit parent process environment.

            arguments:
            environment (dict): a dictionary with environment variables to be set.
            """
            self.new_environment = environment

        def set_creation_flags(self, creation_flags):
            """Set creation flags.

            arguments:
            creation_flags (int): See pywin32 documentation or windows api for documentation.
            """
            self.dw_creation_flags = creation_flags

        def add_limit_flags(self, limit_flags):
            """Add limit flag."""
            new_limit_flags = self.get_limit_flags() | limit_flags
            self.set_limit_flags(new_limit_flags)

        def remove_limit_flags(self, limit_flags):
            """Remove limit flag."""
            new_limit_flags = self.get_limit_flags() & ~limit_flags
            self.set_limit_flags(new_limit_flags)

        def get_limit_flags(self):
            """Get limit flags."""
            return self.limit_info_dict["BasicLimitInformation"]["LimitFlags"]

        def set_limit_flags(self, win32job_limit):
            """Set limit flags.

            arguments:
            win32job_limit (int): See pywin32 documentation or windows api documentation.
            """
            self.limit_info_dict.update({"BasicLimitInformation": {"LimitFlags": win32job_limit}})

        def poll(self):
            """Return process status.

            If process is still running it will return exitcode 259 (STILL_ACTIVE).
            """
            return win32process.GetExitCodeProcess(self.process_handle)

        def terminate(self):
            """Terminate all processes associated with the job."""
            return win32job.TerminateJobObject(self.job_handle)

except ImportError:

    class PtWin32Process:
        """Dummy."""

        def __init__(self, kill_on_job_close=True):
            """Dummy."""
            raise OSError("pywin32 is not installed or is not compatible with this platform")



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for verifying the model yaml files."""

import argparse
import logging
import sys
import typing
from pathlib import Path
from voluptuous import All, MultipleInvalid, Optional, Required, Schema
from ruamel.yaml import YAML
from powertrain_build.interface.application import Application
from powertrain_build.interface.base import BaseApplication

PARSER_HELP = "Verify the model yaml files."


class ModelYmlVerification(BaseApplication):
    """Class for verifying the model yaml files."""

    def __init__(self, base_application):
        self.base_application = base_application
        self.raw = {}
        self.app_insignals = self.get_insignals_name()
        self.app_outsignals = self.get_outsignals_name()
        self.signal_properties = {}
        self.model_name = None
        self.error_printed = False

    def read_translation(self, translation_file):
        """Read specification of the yaml file.

        Args:
            translation_file (Path): file with specs.
        """

        if not translation_file.is_file():
            return {}
        with open(translation_file, encoding="utf-8") as translation:
            try:
                yaml = YAML(typ='safe', pure=True)
                self.raw = yaml.load(translation)
            except yaml.YAMLError as e:
                self.raw = {}
                if hasattr(e, 'problem_mark'):
                    mark = e.problem_mark
                    self.error("Error while reading model file, verification of this file cannot continue until this "
                               f"is fixed:\nFile: {translation_file}\nLine: {mark.line + 1}\nColumn: {mark.column + 1}")
                else:
                    self.error("Error while reading model file, verification of this file cannot continue until this "
                               f"is fixed:\nFile: {translation_file}")

    def validate_signal_schema(self, signal_spec, signal_direction, is_hal, is_service):
        """Validate if signal have a correct schema in model yaml file.

        Args:
            signal_spec (dict): signal specification.
            signal_direction (str): insignal or outsignal.
            is_hal (Bool): signal comming from hal
            is_service (Bool): signal comming from service
        """
        if is_hal:
            signal_schema = Schema({Required('insignal'): All(str), Optional('property'): All(str)})
        elif is_service:
            signal_schema = Schema({Required(signal_direction): All(str), Optional('property'): All(str)})
        else:
            signal_schema = Schema({Required(signal_direction): All(str), Required('property'): All(str)})
        try:
            signal_schema(signal_spec)
        except MultipleInvalid as e:
            self.error(f"{e} in {self.model_name}")

    def validate_group_schema(self, group_spec):
        """Validate if device proxy signal group and hal endpoint
        have correct schema.

        Args:
            group_spec (dict): dp signal group or hal endpoint.
        """
        group_schema = Schema({Required(str): [{Required(str): list}]})
        try:
            group_schema(group_spec)
        except MultipleInvalid as e:
            self.error(self.model_name + ' ' + str(e))

    def validate_service_schema(self, service_spec):
        """Validate if service schema in model yaml file.

        Args:
            service_spec (dict): service in model yaml file.
        """
        service_schema = Schema(
            {
                Required(str): {
                    Optional('properties'): [{Required(str): list}],
                    Optional('methods'): [{Required(str): list}]
                }
            }
        )
        try:
            service_schema(service_spec)
        except MultipleInvalid as e:
            self.error(self.model_name + ' ' + str(e))

    def validate_schema(self):
        """Validate interface schema in model yaml file.

        Interface could be hal, dp(signal/signal group) and serivece.
        """
        schema = Schema({
            Optional('hal'): dict, Optional('signal_groups'): dict,
            Optional('service'): dict, Optional('signals'): dict})
        try:
            schema(self.raw)
        except MultipleInvalid as e:
            self.error(self.model_name + ' ' + str(e))

    def parse_hal_definition(self, hal):
        """Parse hal definition.

        Args:
            hal (dict): hal specification in model yaml file.
        """
        self.parse_group_definitions(hal, is_hal=True)

    def parse_service_definitions(self, service):
        """Parse service.

        Args:
            service (dict): service in model yaml file.
        """
        if service:
            self.validate_service_schema(service)
        for service_name, definition in service.items():
            for endpoints in definition.get('properties', []):
                for endpoint, signals in endpoints.items():
                    self.verify_signals({service_name: signals}, is_service=True, endpoint=endpoint)
            for endpoints in definition.get('methods', []):
                for endpoint, signals in endpoints.items():
                    self.verify_signals({service_name: signals}, is_service=True, endpoint=endpoint)

    def parse_group_definitions(self, signal_groups, is_hal=False):
        """Parse signal groups.

        Args:
            signal_groups (dict): Hal/dp signal group in yaml file.
            is_hal (Bool): hal signal
        """
        if signal_groups:
            self.validate_group_schema(signal_groups)
        for interface, group_definitions in signal_groups.items():
            for group in group_definitions:
                for group_name, signals in group.items():
                    self.verify_signals({interface: signals}, is_hal, group=group_name)

    def verify_signals(self, signals_definition, is_hal=False, is_service=False, endpoint=None, group=None):
        """verify signal in-signal and out-signal in model yaml file.

        Args:
            signals_definition (dict): parsed signals in model yaml file.
            is_hal (Bool): hal signal.
            is_service (Bool): service signal.
            endpoint (str): service endpoint.
            group(str): hal group.
        """
        for interface, specifications in signals_definition.items():
            for specification in specifications:
                in_out_signal = [key for key in specification.keys() if 'signal' in key]
                signal_name = specification[in_out_signal[0]]
                self.validate_signal_schema(specification, in_out_signal[0], is_hal, is_service)
                if in_out_signal == []:
                    self.error(f"signal is not defined for {interface} in {self.model_name}!")
                if 'in' not in in_out_signal[0] and 'out' not in in_out_signal[0]:
                    self.error(f"in and out signal must be added to signal {specification['signal']}")
                elif 'in' in in_out_signal[0] and specification[in_out_signal[0]] not in self.app_insignals:
                    self.error(
                        f"{specification['insignal']} is not defined as an insignal in json file")
                elif "out" in in_out_signal[0] and specification[in_out_signal[0]] not in self.app_outsignals:
                    self.error(
                        f"{specification['outsignal']} is not defined as an outsignal in json file")
                else:
                    if is_service:
                        if specification.get('property') is None:
                            self.verify_primitive(f"{interface}.{endpoint}.{signal_name}", in_out_signal[0])
                        else:
                            self.verify_primitive(
                                f"{interface}.{endpoint}.{specification['property']}.{signal_name}",
                                in_out_signal[0])
                    elif is_hal:
                        if specification.get('property') is None:
                            self.verify_primitive(f"{interface}.{group}.{signal_name}", in_out_signal[0])
                        else:
                            self.verify_primitive(f"{interface}.{group}.{specification['property']}.{signal_name}",
                                                  in_out_signal[0])
                    else:
                        self.verify_primitive(f"{interface}.{specification['property']}.{signal_name}",
                                              in_out_signal[0])

    def check_duplicate_signals(self, property_spec, in_out):
        """Check if each signal appears only once for each model.
        It is ok for two insignals with the same name to be mapped to the same primitive
        if they are in different models.
        It is not ok for a outsignal to map to the same interface twice, but it is ok to map to
        different interfaces (with same or different property name).

        Args:
            property_spec (str): property specification.
            in_out (str): whether it is an in- or outsignal.
        """
        signal_name = property_spec.split('.')[-1]
        interface_name = property_spec.split('.')[0]
        for model, spec in self.signal_properties.items():
            for primitive in spec:
                if signal_name in primitive.split('.'):
                    if property_spec not in primitive:
                        if "in" in in_out:
                            self.error(
                                f"You can't connect a signal {signal_name} in {self.model_name} model to two "
                                f"different primitives. It's already connected in {model} model")
                        else:
                            if interface_name in primitive.split('.'):
                                self.error(
                                    f"You can't connect a signal {signal_name} in {self.model_name} model "
                                    f"to the same interface ({interface_name}) twice. "
                                    f"It's already connected as {primitive} in model {model}.")

                    else:
                        if model == self.model_name:
                            self.error(f"You can't connect signal {signal_name} in {self.model_name} model twice.")
                        elif "out" in in_out:
                            self.error(
                                f"You can't connect signal {signal_name} in {self.model_name} model to the same "
                                f"primitive as in another model. It is already defined in {model}")

    def check_property(self, property_spec):
        """Check if we have only one signal written for the same property.

        Args:
            property_spec (str): property specification.
        """
        signal_name = property_spec.split('.')[-1]
        for model, spec in self.signal_properties.items():
            for primitive in spec:
                if ('.'.join(property_spec.split('.')[:-1]) == '.'.join(primitive.split('.')[:-1])
                        and signal_name != primitive.split('.')[-1]):
                    self.error(
                            f"You can't connect another signal to the existing property {property_spec} in "
                            f"{self.model_name} model, because it is already defined in {model} model.")

    def verify_primitive(self, property_spec, in_out):
        """Runs the necessary tests.

        Args:
            property_spec (str): property specification.
            in_out (str): whether it is an in- or outsignal.
        """
        self.check_duplicate_signals(property_spec, in_out)
        self.check_property(property_spec)
        self.signal_properties[self.model_name].append(property_spec)

    def get_insignals_name(self):
        """Base application in-signals.
        """
        app_insignals = []
        for signal in self.base_application.insignals:
            app_insignals.append(signal.name)
        return app_insignals

    def get_outsignals_name(self):
        """Base application out-signals.
        """
        app_outsignals = []
        for signal in self.base_application.outsignals:
            app_outsignals.append(signal.name)
        return app_outsignals

    def parse_definition(self, definition):
        """Parses all definition files

        Args:
            definition (list(Path)): model yaml files.
        """
        for translation in definition:
            path = Path(translation)
            self.model_name = path.name.replace(".yaml", "")
            self.signal_properties[self.model_name] = []
            self.read_translation(translation)
            self.validate_schema()
            self.verify_signals(self.raw.get("signals", {}))
            self.parse_group_definitions(self.raw.get("signal_groups", {}))
            self.parse_hal_definition(self.raw.get("hal", {}))
            self.parse_service_definitions(self.raw.get("service", {}))

    def error(self, msg):
        """ Prints an error message to the terminal.

        Args:
            msg (string): The message to be printed.
        """
        self.error_printed = True
        logging.error(f"{msg}\n")

    def print_success_msg(self):
        """ Prints a success message if no error messages have been printed.
        """
        if not self.error_printed:
            print('Yaml verification done without any errors.')


def get_app(project_config):
    """ Get an app specification for the current project.

    Args:
        config (pathlib.Path): Path to the ProjectCfg.json.
    Returns:
        app (Application): powertrain-build project.
    """
    app = Application()
    app.parse_definition(project_config)
    return app


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the argument parser."""
    parser.add_argument("config", help="The SPA2 project config file", type=Path)
    parser.set_defaults(func=model_yaml_verification_cli)


def model_yaml_verification_cli(args: argparse.Namespace):
    """CLI function for model yaml verification."""
    app = get_app(args.config)
    model_yamls = app.get_translation_files()
    model_yaml_ver = ModelYmlVerification(app)
    model_yaml_ver.parse_definition(model_yamls)
    model_yaml_ver.print_success_msg()


def main(argv: typing.Optional[typing.List[str]] = None):
    """Main function for model yaml verification."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for providing an xlrd interface for csv files.

TODO: Handle sheet names in a better way!
"""

import csv
import os
import re


class WorkBook:
    """Emulate xlrd of csv files."""

    def __init__(self, csv_files):
        """Init."""
        self._worksheets = {}
        self._read_csv_files_to_worksheets(csv_files)

    def _read_csv_files_to_worksheets(self, csv_files):
        """Read CSV-files, and store them as worksheets."""
        csv_files = [csv_files] if isinstance(csv_files, str) else csv_files
        if len(csv_files) > 1:
            for file_ in csv_files:
                sheet_name = re.match(r'.*?\w*_([\w\$]+)\.\w+', file_).group(1)
                self._worksheets[sheet_name] = WorkSheet(sheet_name, file_)
        else:
            _, sheet_name = os.path.split(csv_files[0])
            self._worksheets[sheet_name] = WorkSheet(sheet_name, csv_files[0])

    def sheet_by_name(self, name):
        """Get a worksheet name from a workbook class."""
        return self._worksheets[name]

    def single_sheet(self):
        """Get single worksheet if only one file was read."""
        return next(iter(self._worksheets.values())) if len(self._worksheets) == 1 else None


class WorkSheet:
    """Emulate xlrd for csv files."""

    def __init__(self, sheet_name, file_name):
        """Init."""
        self._sheet_name = sheet_name
        self._read_work_sheet(file_name)

    def _read_work_sheet(self, file_name):
        """Read csv-file as a worksheet."""
        with open(file_name, 'r', encoding="ISO-8859-1") as csvfile:
            reader = csv.reader(csvfile, delimiter=';')
            self._rows = [[Data(elem) for elem in row] for row in reader]
            self._nrows = len(self._rows)

    @property
    def nrows(self):
        """Get number of rows in a worksheet."""
        return self._nrows

    def row(self, row):
        """Get row from sheet."""
        return self._rows[row]


class Data:
    """Class for wrapping data in a worksheet."""

    def __init__(self, value):
        """Init."""
        # TODO: Investigate quoting string values when exporting to CSV
        if re.fullmatch('[+-]?[0-9]+', value):
            self._value = int(value)
        elif re.fullmatch('[+-]?[0-9.]+', value):
            self._value = float(value)
        else:
            self._value = value

    def __repr__(self):
        """Get representation of object."""
        return self._value

    @property
    def value(self):
        """Provide the value of the data object."""
        return self._value



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for generating core API dummy code for Bosch and Denso suppliers.

These files are needed for building test SW before the supplier has
updated the core with all the needed core id definitions.
"""

import os

from powertrain_build import build_defs
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import get_ec_type
from powertrain_build.unit_configs import CodeGenerators


class CoreDummy(ProblemLogger):
    """A class for core API dummy file generation."""

    def __init__(self, core_cfg, unit_cfg):
        """Initialize with core API configuration.

        Args:
            core_cfg (Core): Configuration object
        """
        super().__init__()
        self._core_cfg = core_cfg
        self._unit_cfg = unit_cfg
        self._uint16_type = self._get_uint_16_type()
        self.fh_c = None
        self.fh_h = None

    def _get_uint_16_type(self):
        if CodeGenerators.target_link in self._unit_cfg.code_generators:
            return 'UInt16'
        return get_ec_type('UInt16')

    def _gen_rb_header(self, file_name):
        """Generate the RB start of the c and h files for the dummy core definition files.

        Args:
            file_name : The RB header file
        """
        self.fh_c.write('/* Autogenerated core id dummy file */\n\n')
        self.fh_c.write(f'#include "{file_name}"\n\n')
        self.fh_c.write(f'#include "{build_defs.CVC_CODE_START}"\n\n')
        self.fh_c.write('#ifndef VcEvDummy_ID\n')
        self.fh_c.write('  #define VcEvDummy_ID 1\n')
        self.fh_c.write('#endif /* VcEvDummy_ID */\n')
        self.fh_c.write('#ifndef VcEvDummy_DEBMODE\n')
        self.fh_c.write('  #define VcEvDummy_DEBMODE 0\n')
        self.fh_c.write('#endif /* VcEvDummy_DEBMODE */\n')
        self.fh_c.write('#ifndef VcFiDummy_ID\n')
        self.fh_c.write('  #define VcFiDummy_ID 1\n')
        self.fh_c.write('#endif /* VcFiDummy_ID */\n')
        self.fh_c.write('#ifndef Vc06Dummy_ID\n')
        self.fh_c.write('  #define Vc06Dummy_ID 1\n')
        self.fh_c.write('#endif /* Vc06Dummy_ID */\n')
        self.fh_c.write('#ifndef Vc09Dummy_ID\n')
        self.fh_c.write('  #define Vc09Dummy_ID 1\n')
        self.fh_c.write('#endif /* Vc09Dummy_ID */\n')
        self.fh_c.write('#ifndef VcRvDummy_ID\n')
        self.fh_c.write('  #define VcRvDummy_ID 1\n')
        self.fh_c.write('#endif /* VcRvDummy_ID */\n\n')

        def_name = file_name.replace('.', '_').upper()
        self.fh_h.write('/* Autogenerated core id dummy file */\n\n')
        self.fh_h.write(f'#ifndef {def_name}\n')
        self.fh_h.write(f'#define {def_name}\n')
        self.fh_h.write('#include "vcc_rb_core.h"\n')
        self.fh_h.write(self._unit_cfg.base_types_headers)

    def _gen_rb_event_dummies(self):
        """Generate RB dummy event declarations."""
        for id_, comment in self._core_cfg.get('EventIDs', {}).items():
            self.fh_h.write(f'#ifndef {id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   extern const DSM_DFCType DFC_{id_} ;\n')
            self.fh_h.write(f'   #define {id_} (((DFC_{id_}.debmode) << 12u) | '
                            f'(DFC_{id_}.id))\n')
            self.fh_h.write(f'   #define {id_}_Dummy\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifdef {id_}_Dummy\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const DSM_DFCType DFC_{id_} = '
                            '{VcEvDummy_ID, VcEvDummy_DEBMODE};\n')
            self.fh_c.write(f'   #warning "CoreID: {id_} is not defined in '
                            'the supplier SW"\n')
            self.fh_c.write('#endif\n\n')

    def _gen_rb_fid_dummies(self):
        """Generate RB dummy function id declarations."""
        for id_, comment in self._core_cfg.get('FunctionIDs', {}).items():
            self.fh_h.write(f'#ifndef {id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   extern const DSM_FIdType FId_{id_} ;\n')
            self.fh_h.write(f'   #define {id_} (FId_{id_}.id)\n')
            self.fh_h.write(f'   #define {id_}_Dummy\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifdef {id_}_Dummy\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const DSM_FIdType FId_{id_} = {{VcFiDummy_ID}};\n')
            self.fh_c.write(f'   #warning "CoreID: {id_} is not defined in the supplier SW"\n')
            self.fh_c.write('#endif\n\n')

    def _gen_rb_iumpr_dummies(self):
        """Generate RB dummy IUMPR declarations."""
        for id_, comment in self._core_cfg.get('IUMPR', {}).items():
            self.fh_h.write(f'#ifndef {id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   extern const DSM_FIdType FId_{id_} ;\n')
            self.fh_h.write(f'   #define {id_} (FId_{id_}.id)\n')
            self.fh_h.write(f'   #define {id_}_Dummy\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifdef {id_}_Dummy\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const DSM_FIdType FId_{id_} = {{Vc09Dummy_ID}};\n')
            self.fh_c.write(f'   #warning "CoreID: {id_} is not defined in the supplier SW"\n')
            self.fh_c.write('#endif\n\n')

    def _gen_rb_mode06_dummies(self):
        """Generate RB dummy mode$06 declarations."""
        for id_, comment in self._core_cfg.get('Mode$06', {}).items():
            self.fh_h.write(f'#ifndef {id_}\n')
            self.fh_h.write(f'   /* {comment[0]}, UAS {comment[1]} */\n')
            self.fh_h.write(f'   extern const DSM_DTRType DTR_{id_};\n')
            self.fh_h.write(f'   #define {id_} (DTR_{id_}.id)\n')
            self.fh_h.write(f'   #define {id_}_Dummy\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifdef {id_}_Dummy\n')
            self.fh_c.write(f'   /* {comment[0]}, UAS {comment[1]} */\n')
            self.fh_c.write(f'   const DSM_DTRType DTR_{id_} = {{Vc06Dummy_ID}};\n')
            self.fh_c.write(f'#warning "CoreID: {id_} is not defined in the supplier SW"\n')
            self.fh_c.write('#endif\n\n')

    def _gen_rb_rnk_dummies(self):
        """Generate RB dummy ranking declarations."""
        for id_, comment in self._core_cfg.get('Ranking', {}).items():
            self.fh_h.write(f'#ifndef {id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   extern const {self._uint16_type} DSMAppl_RnkValStorg_RnkId_{id_};\n')
            self.fh_h.write(f'   #define {id_} DSMAppl_RnkValStorg_RnkId_{id_}\n')
            self.fh_h.write(f'   #define {id_}_Dummy\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifdef {id_}_Dummy\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const {self._uint16_type} DSMAppl_RnkValStorg_RnkId_{id_} = '
                            'VcRvDummy_ID;\n')
            self.fh_c.write(f'   #warning "CoreID: {id_} is not defined in the supplier SW"\n')
            self.fh_c.write('#endif\n\n')

    def _gen_rb_end(self, file_name):
        """Generate RB footer of c and h files for dummy core definition files."""
        def_name = file_name.replace('.', '_').upper()

        self.fh_h.write(f'#endif /* {def_name} */\n')
        self.fh_h.write('/*-------------------------------------'
                        '---------------------------------------*\\\n')
        self.fh_h.write('  END OF FILE\n')
        self.fh_h.write('\\*-------------------------------------'
                        '---------------------------------------*/\n')

        self.fh_c.write(f'#include "{build_defs.CVC_CODE_END}"\n')
        self.fh_c.write('/*--------------------------------------'
                        '--------------------------------------*\\\n')
        self.fh_c.write('  END OF FILE\n')
        self.fh_c.write('\\*--------------------------------------'
                        '--------------------------------------*/\n')

    def generate_rb_core_dummy_files(self, file_name):
        """Generate core API dummy files for Bosch projects."""
        cname = file_name + '.c'
        hname = file_name + '.h'
        with open(cname, 'w', encoding="utf-8") as self.fh_c:
            with open(hname, 'w', encoding="utf-8") as self.fh_h:
                _, f_name = os.path.split(hname)
                self._gen_rb_header(f_name)
                self._gen_rb_event_dummies()
                self._gen_rb_fid_dummies()
                self._gen_rb_iumpr_dummies()
                self._gen_rb_mode06_dummies()
                self._gen_rb_rnk_dummies()
                self._gen_rb_end(f_name)

    def _gen_dg2_header(self, file_name):
        """Generate Denso Gen2/3 header of c and h files for dummy core definition files."""
        self.fh_c.write('/* Autogenerated core id dummy file */\n\n')
        self.fh_c.write(f'#include "{file_name}"\n')
        self.fh_c.write(self._unit_cfg.base_types_headers)
        self.fh_c.write(f'#include "{build_defs.CVC_CODE_START}"\n\n')
        self.fh_c.write('#define Vc06Undef 1\n')
        self.fh_c.write('#define Vc09Undef 1\n')
        self.fh_c.write('#define VcRvUndef 1\n\n')
        self.fh_h.write('/* Autogenerated core id dummy file */\n\n')
        def_name = file_name.replace('.', '_').upper()
        self.fh_h.write(f'#ifndef {def_name}\n')
        self.fh_h.write(f'#define {def_name}\n')
        self.fh_h.write('\n#include "VcCoreSupplierAbstraction.h"\n\n')
        self.fh_h.write('/* Check if denso make env. has defined the AUTOSAR declarations\n')
        self.fh_h.write('   if not, do dummy declaration to be able to build a dummy I/F in\n')
        self.fh_h.write('   an old make environment */\n')
        self.fh_h.write('#ifndef FUNC\n')
        self.fh_h.write('  #define FUNC(rettype, memclass) rettype /* from Compiler.h */\n')
        self.fh_h.write('  #define P2VAR(ptrtype, memclass, ptrclass) ptrtype * '
                        '/* from Compiler.h */\n')
        self.fh_h.write('  typedef unsigned char Std_ReturnType; /* from Std_Types.h */\n')
        self.fh_h.write('  typedef unsigned char Dem_EventStatusType;\n')
        self.fh_h.write('  typedef unsigned char boolean;\n')
        self.fh_h.write('  #define AUTOMATIC\n')
        self.fh_h.write('  #define RTE_APPL_DATA\n')
        self.fh_h.write('  #define RTE_CODE /* from Compiler_cfg.h */\n')
        self.fh_h.write(f'#endif /* {def_name} */\n\n')
        self.fh_h.write(f'#include "{build_defs.CVC_CODE_START}"\n')

    def _gen_dg2_event_dummies(self):
        """Generate the Denso Gen2+ dummy event declarations."""
        for id_, comment in self._core_cfg.get('EventIDs', {}).items():
            # Compare with core0/app/OBDFW/bdcore/Rte_Diag.h
            self.fh_h.write(f'#ifndef Rte_Call_Event_{id_}_SetEventStatus\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write('   extern FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_Event_{id_}_SetEventStatus '
                            '(Dem_EventStatusType EventStatus);\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifndef Rte_Call_Event_{id_}_SetEventStatus\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write('   FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_Event_{id_}_SetEventStatus '
                            '(Dem_EventStatusType EventStatus) { return 0;}\n')
            self.fh_c.write('#endif\n\n')
            # Generate Dem_SetEventDisabled dummmies
            self.fh_h.write(f'#ifndef Rte_Call_Event_{id_}_SetEventDisabled\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write('   extern FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_Event_{id_}_SetEventDisabled (void);\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifndef Rte_Call_Event_{id_}_SetEventDisabled\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write('   FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_Event_{id_}_SetEventDisabled '
                            '(void) { return 0;}\n')
            self.fh_c.write('#endif\n\n')

    def _gen_dg2_fid_dummies(self):
        """Generate the Denso Gen2+ dummy function id declarations."""
        for id_, comment in self._core_cfg.get('FunctionIDs', {}).items():
            self.fh_h.write(f'#ifndef Rte_Call_FI_{id_}_GetFunctionPermission\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write('   extern FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_FI_{id_}_GetFunctionPermission '
                            '(P2VAR(boolean, AUTOMATIC, RTE_APPL_DATA) Permission);\n')
            self.fh_h.write('#endif\n\n')

            self.fh_c.write(f'#ifndef Rte_Call_FI_{id_}_GetFunctionPermission\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write('   FUNC(Std_ReturnType, RTE_CODE) '
                            f'Rte_Call_Diag_FI_{id_}_GetFunctionPermission '
                            '(P2VAR(boolean, AUTOMATIC, RTE_APPL_DATA) Permission) '
                            '{ *Permission = 1; return 0;}\n')
            self.fh_c.write('#endif\n\n')

    def _gen_dg2_iumpr_dummies(self):
        """Generate the Denso Gen2+ dummy IUMPR declarations."""
        for id_, comment in self._core_cfg.get('IUMPR', {}).items():
            self.fh_h.write(f'#ifndef IUMID_{id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   #define IUMID_{id_} cIUMID_{id_}\n')
            self.fh_h.write(f'   #define {id_}_DUMMY\n')
            self.fh_h.write(f'   extern const {self._uint16_type} cIUMID_{id_};\n')
            self.fh_h.write('#endif\n\n')
            self.fh_c.write(f'#ifdef {id_}_DUMMY\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const {self._uint16_type} cIUMID_{id_} = Vc09Undef;\n')
            self.fh_c.write('#endif\n\n')

    def _gen_dg2_mode06_dummies(self):
        """Generate the Denso Gen2+ dummy mode$06 declarations."""
        for id_, comment in self._core_cfg.get('Mode$06', {}).items():
            self.fh_h.write(f'#ifndef TR_{id_}\n')
            self.fh_h.write(f'   /* {comment[0]}, UAS {comment[1]} */\n')
            self.fh_h.write(f'   #define TR_{id_} cTR_{id_}\n')
            self.fh_h.write(f'   #define {id_}_DUMMY\n')
            self.fh_h.write(f'   extern const {self._uint16_type} cTR_{id_};\n')
            self.fh_h.write('#endif\n\n')
            self.fh_c.write(f'#ifdef {id_}_DUMMY\n')
            self.fh_c.write(f'   /* {comment[0]}, UAS {comment[1]} */\n')
            self.fh_c.write(f'   const {self._uint16_type} cTR_{id_} = Vc06Undef;\n')
            self.fh_c.write('#endif\n\n')

    def _gen_dg2_rnk_dummies(self):
        """Generate the Denso Gen2+ dummy ranking declarations."""
        for id_, comment in self._core_cfg.get('Ranking', {}).items():
            self.fh_h.write(f'#ifndef RVID_{id_}\n')
            self.fh_h.write(f'   /* {comment} */\n')
            self.fh_h.write(f'   #define RVID_{id_} cRVID_{id_}\n')
            self.fh_h.write(f'   #define {id_}_DUMMY\n')
            self.fh_h.write(f'   extern const {self._uint16_type} cRVID_{id_};\n')
            self.fh_h.write('#endif\n\n')
            self.fh_c.write(f'#ifdef {id_}_DUMMY\n')
            self.fh_c.write(f'   /* {comment} */\n')
            self.fh_c.write(f'   const {self._uint16_type} cRVID_{id_} = VcRvUndef;\n')
            self.fh_c.write('#endif\n\n')

    def _gen_dg2_end(self, file_name):
        """Generate Denso Gen2+ footer of c and h files for dummy core definition files."""
        def_name = file_name.replace('.', '_').upper()
        self.fh_h.write(f'#include "{build_defs.CVC_CODE_END}"\n')
        self.fh_h.write(f'#endif /* {def_name} */\n')
        self.fh_h.write('/*------------------------------------------'
                        '----------------------------------*\\\n')
        self.fh_h.write('  END OF FILE\n')
        self.fh_h.write('\\*-----------------------------------------'
                        '-----------------------------------*/\n')
        self.fh_c.write(f'#include "{build_defs.CVC_CODE_END}"\n')
        self.fh_c.write('/*------------------------------------------'
                        '----------------------------------*\\\n')
        self.fh_c.write('  END OF FILE\n')
        self.fh_c.write('\\*-----------------------------------------'
                        '-----------------------------------*/\n')

    def generate_dg2_core_dummy_files(self, file_name):
        """Generate the core API dummy files for Denso gen2+ projects."""
        cname = file_name + '.c'
        hname = file_name + '.h'
        with open(cname, 'w', encoding="utf-8") as self.fh_c:
            with open(hname, 'w', encoding="utf-8") as self.fh_h:
                _, f_name = os.path.split(hname)
                self._gen_dg2_header(f_name)
                self._gen_dg2_event_dummies()
                self._gen_dg2_fid_dummies()
                self._gen_dg2_iumpr_dummies()
                self._gen_dg2_mode06_dummies()
                self._gen_dg2_rnk_dummies()
                self._gen_dg2_end(f_name)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for model out-port interfaces."""
import os
from operator import itemgetter
import math

import powertrain_build.build_defs as bd
from powertrain_build import signal
from powertrain_build.a2l import A2l
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import byte_size, get_bitmask


class DummySpm(ProblemLogger):
    """Generate c-files which defines missing outport variables in the model out-port interface.

    The models declare all in-ports as 'external' and powertrain-build will then
    generate any missing outports in the correct #if/#endif guards here.
    One c-file per outport origin model should be generated.

    * Generate c-code from matlab/TL script with #if/#endif guards.
      - Pro: Is generated at the same time as other code and placed in model src folder.
      - Pro: Generic code. Will be cached together with TL-generated code.
      - Con: m-script
    * Generate c-code from python with only needed variables. No preprocessor directives.
      - Pro: Python
      - Pro: Simpler c-file with only used variables.
      - Con: Not generic! Not cached?

    """

    __asil_level_map = {
        'A': (bd.CVC_DISP_ASIL_A_START, bd.CVC_DISP_ASIL_A_END),
        'B': (bd.CVC_DISP_ASIL_B_START, bd.CVC_DISP_ASIL_B_END),
        'C': (bd.CVC_DISP_ASIL_C_START, bd.CVC_DISP_ASIL_C_END),
        'D': (bd.CVC_DISP_ASIL_D_START, bd.CVC_DISP_ASIL_D_END),
        'QM': (bd.CVC_DISP_START, bd.CVC_DISP_END)}

    def __init__(self, missing_outports, prj_cfg, feature_cfg, unit_cfg, user_defined_types, basename):
        """Constructor.

        Args:
            missing_outports (list): undefined outports based on unit config variables.
            prj_cfg (BuildProjConfig): Build project class holding where files should be stored.
            feature_cfg (FeatureConfig): Feature configs from SPM_Codeswitch_Setup.
            unit_cfg (UnitConfigs): Class holding all unit interfaces.
            user_defined_types (UserDefinedTypes): Class holding user defined data types.
            basename (str): the basename of the outvar, used for .c and .a2l creation.

        See :doc:`Unit config <unit_config>` for information on the 'outport' dict.
        """
        super().__init__()
        self._prj_cfg = prj_cfg
        self._feature_cfg = feature_cfg
        self._unit_cfg = unit_cfg
        self._enumerations = user_defined_types.get_enumerations()
        self._common_header_files = user_defined_types.common_header_files
        self._name = basename
        self.use_volatile_globals = prj_cfg.get_use_volatile_globals()
        self._missing_outports = self._restruct_input_data(missing_outports)

    def _get_byte_size(self, data_type):
        """Get byte size of a data type.
        Enumeration byte sizes are derived from the underlying data type.

        Args:
            data_type (str): Data type.
        Returns:
            byte_size(powertrain_build.types.byte_size): Return result of powertrain_build.types.byte_size.
        """
        if data_type in self._enumerations:
            return byte_size(self._enumerations[data_type]['underlying_data_type'])
        return byte_size(data_type)

    def _restruct_input_data(self, outports):
        """Restructure all the input variables per data-type."""
        outports.sort(key=itemgetter('name'))
        outports.sort(key=lambda var: self._get_byte_size(var['type']), reverse=True)
        new_outports = {}
        for outport in outports:
            integrity_level = outport.get('integrity_level', 'QM')
            if integrity_level == 'QM':
                outport['cvc_type'] = 'CVC_DISP'
            else:
                outport['cvc_type'] = f'ASIL_{integrity_level}/CVC_DISP_ASIL_{integrity_level}'
            if integrity_level in new_outports:
                new_outports[integrity_level].append(outport)
            else:
                new_outports.update({integrity_level: [outport]})
        return new_outports

    def _a2l_dict(self, outports):
        """Return a dict defining all parameters for a2l-generation."""
        res = {
            'vars': {},
            'function': 'VcDummy_spm'
        }

        for outport in [port for sublist in outports.values() for port in sublist]:
            var = outport['name']

            if outport['type'] in self._enumerations:
                data_type = self._enumerations[outport['type']]['underlying_data_type']
            else:
                data_type = outport['type']

            resv = res['vars']
            resv.setdefault(var, {})['a2l_data'] = {
                    'bitmask': get_bitmask(data_type),
                    'description': outport.get('description', ''),
                    'lsb': '2^{}'.format(int(math.log2(outport.get('lsb', 1)))
                                         if outport.get('lsb') not in ['-', '']
                                         else 0),
                    'max': outport.get('max'),
                    'min': outport.get('min'),
                    'offset': -(outport.get('offset', 0) if outport.get('offset') not in ['-', ''] else 0),
                    'unit': outport['unit'],
                    'x_axis': None,
                    'y_axis': None}
            resv[var]['function'] = ['VcEc']
            resv[var]['var'] = {'cvc_type': outport['cvc_type'],
                                'type': data_type,
                                'var': var}
            resv[var]['array'] = outport.get('width', 1)
            res.update({'vars': resv})
        return res

    def generate_code_files(self, dst_dir):
        """Generate code and header files.

        Args:
            dst_dir (str): Path to destination directory.

        """
        h_file_path = os.path.join(dst_dir, f'{self._name}.h')
        self._generate_h_file(h_file_path, self._missing_outports)
        c_file_path = os.path.join(dst_dir, f'{self._name}.c')
        self._generate_c_file(c_file_path, self._missing_outports)

    def generate_a2l_files(self, dst_dir):
        """Generate A2L files.

        Args:
            dst_dir (str): Path to destination directory.

        """
        filename = f"{os.path.join(dst_dir, self._name)}.a2l"
        a2l_dict = self._a2l_dict(self._missing_outports)
        a2l = A2l(a2l_dict, self._prj_cfg)
        a2l.gen_a2l(filename)

    def _generate_h_file(self, file_path, outports):
        """Generate header file.

        Args:
            file_path (str): File path to generate.
        """
        file_name = os.path.basename(file_path).split('.')[0]
        with open(file_path, 'w', encoding="utf-8") as fh:
            fh.write(f'#ifndef {file_name.upper()}_H\n')
            fh.write(f'#define {file_name.upper()}_H\n')

            fh.write(self._unit_cfg.base_types_headers)
            fh.write('#include "VcCodeSwDefines.h"\n')
            for common_header_file in self._common_header_files:
                fh.write(f'#include "{common_header_file}"\n')

            for integrity_level in outports.keys():
                disp_start = bd.PREDECL_ASIL_LEVEL_MAP[integrity_level]['DISP']['START']
                disp_end = bd.PREDECL_ASIL_LEVEL_MAP[integrity_level]['DISP']['END']
                fh.write('\n')
                fh.write(f'#include "{disp_start}"\n')
                for outport in outports[integrity_level]:
                    if outport['class'] not in signal.INPORT_CLASSES:
                        self.warning(f'inport {outport["name"]} class {outport["class"]} is not an inport class')

                    array = ''
                    width = outport['width']
                    if not isinstance(width, list):
                        width = [width]
                    if len(width) != 1 or width[0] != 1:
                        for w in width:
                            if w > 1:
                                if not isinstance(w, int):
                                    self.critical(f'{outport["name"]} widths must be integers. Got "{type(w)}"')
                                array += f'[{w}]'
                            elif w < 0:
                                self.critical(f'{outport["name"]} widths can not be negative. Got "{w}"')
                    if self.use_volatile_globals:
                        fh.write(f"extern volatile {outport['type']} {outport['name']}{array};\n")
                    else:
                        fh.write(f"extern {outport['type']} {outport['name']}{array};\n")
                fh.write(f'#include "{disp_end}"\n')

            fh.write(f'#endif /* {file_name.upper()}_H */\n')

    def _generate_c_file(self, file_path, outports):
        """Generate C-file for inports that are missing outports except for supplier ports."""
        file_name = os.path.basename(file_path).split('.')[0]
        base_header = f'#include "{file_name}.h"\n'

        with open(file_path, 'w', encoding="utf-8") as fh_c:
            fh_c.write(base_header)
            for integrity_level in outports.keys():
                disp_start = bd.CVC_ASIL_LEVEL_MAP[integrity_level]['DISP']['START']
                disp_end = bd.CVC_ASIL_LEVEL_MAP[integrity_level]['DISP']['END']
                fh_c.write('\n')
                fh_c.write(f'#include "{disp_start}"\n')
                for outport in outports[integrity_level]:
                    if outport['class'] not in signal.INPORT_CLASSES:
                        self.warning(f'inport {outport["name"]} class {outport["class"]} is not an inport class')

                    width = outport['width']
                    if outport['type'] in self._enumerations:
                        if self._enumerations[outport['type']]['default_value'] is not None:
                            init_value = self._enumerations[outport['type']]['default_value']
                        else:
                            self.warning('Initializing enumeration %s to "zero".', outport['type'])
                            init_value = [
                                k for k, v in self._enumerations[outport['type']]['members'].items() if v == 0
                            ][0]
                        if width != 1:
                            self.critical(f'{outport["name"]} enumeration width must be 1. Got "{width}"')
                        fh_c.write(f"{outport['type']} {outport['name']} = {init_value};\n")
                    else:
                        if not isinstance(width, list):
                            width = [width]
                        if len(width) == 1 and width[0] == 1:
                            array = ' = 0'
                        else:
                            array = ''
                            for w in width:
                                if w > 1:
                                    if not isinstance(w, int):
                                        msg = f'{outport["name"]} widths must be integers. Got "{type(w)}"'
                                        self.critical(msg)
                                    array += f'[{w}]'
                                elif w < 0:
                                    self.critical(f'{outport["name"]} widths can not be negative. Got "{w}"')
                        if self.use_volatile_globals:
                            fh_c.write(f'volatile {outport["type"]} {outport["name"]}{array};\n')
                        else:
                            fh_c.write(f'{outport["type"]} {outport["name"]}{array};\n')
                fh_c.write(f'#include "{disp_end}"\n')

    def generate_files(self, dst_dir):
        """Generate the files for defining all missing input variables."""
        self.generate_code_files(dst_dir)
        self.generate_a2l_files(dst_dir)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for logging."""
import os
import sys
import logging

LEVEL_NOTSET = 'NOTSET'
LEVEL_DEBUG = 'DEBUG'
LEVEL_INFO = 'INFO'
LEVEL_WARNING = 'WARNING'
LEVEL_ERROR = 'ERROR'
LEVEL_CRITICAL = 'CRITICAL'

LEVELS = {
    LEVEL_NOTSET: logging.NOTSET,
    LEVEL_DEBUG: logging.DEBUG,
    LEVEL_INFO: logging.INFO,
    LEVEL_WARNING: logging.WARNING,
    LEVEL_ERROR: logging.ERROR,
    LEVEL_CRITICAL: logging.CRITICAL
}


def parse_log_level(log_level_name):
    """Convert textual log_level_name to numerical ones defined in logging module."""
    level = log_level_name.upper()
    if level not in LEVELS:
        print(f'Log level "{log_level_name}" invalid, valid list: {", ".join(LEVELS.keys())}', file=sys.stderr)
        level = LEVEL_DEBUG
    return LEVELS[level]


def create_logger(name, handler=None, log_format=None):
    """Create a logger.

    If the handler already have a log format, it will be replaced.

    Args:
        name (str): Name of the logger
        handler (obj): Handler for the logger. Default used if not supplied.
        log_format (str): Format for the handler. Default used if not supplied.'
    Returns:
        logger (obj): A logger with a handler and log format
    """
    new_logger = logging.getLogger(name)
    new_logger.setLevel(parse_log_level(os.getenv('LOG_LEVEL', LEVEL_INFO)))
    if handler is None:
        handler = logging.StreamHandler(sys.stdout)
    if log_format is None:
        log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    handler.setFormatter(logging.Formatter(log_format))
    new_logger.addHandler(handler)
    return new_logger



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for merging of a2l-files."""

import json
import os
import re
from string import Template

from powertrain_build.lib.helper_functions import deep_dict_update
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.a2l_templates import A2lProjectTemplate, A2lSilverTemplate


class A2lMerge(ProblemLogger):
    """Class for merging of a2l-files."""

    def __init__(self, prj_cfg, ucfg, a2l_files_unit, a2l_files_gen):
        """Merge a2l-files based on provided project configuration.

        Removes symbols not included in the projects unit-config files.

        Args:
            prj_cfg (obj): Project config.
            ucfg (obj): Unit config.
            a2l_files_unit (list of str): Files to merge.
            a2l_files_gen (list of str): Files to merge.
        """
        super().__init__()
        self._prj_cfg = prj_cfg
        self._unit_cfg = ucfg
        self._per_unit_cfg = ucfg.get_per_unit_cfg()
        # generate the a2l string
        self._blks = {}
        self._removed_symbols = []
        self.a2l = ""

        # ----- Example blocks in a2l (TargetLink) -----
        #
        # /begin CHARACTERISTIC
        #     cVc_B_SeriesHev	/* Name */
        #     "Series hybrid"	/* LongIdentifier */
        #     VALUE	/* Type */
        #     0x00000000    /* address: cVc_B_SeriesHev */
        #     UBYTE_COL_DIRECT	/* Deposit */
        #     0	/* MaxDiff */
        #     Scaling_3	/* Conversion */
        #     0	/* LowerLimit */
        #     1	/* UpperLimit */
        # /end CHARACTERISTIC
        #
        #  Example of Bosch-nvm signal in nvm:
        #
        # /begin MEASUREMENT
        #     nvm_list_32._sVcDclVu_D_Markow /* Name */
        #     "No description given" /* LongIdentifier */
        #     ULONG   /* Datatype */
        #     VcNvm_1_0_None   /* Conversion */
        #     1   /* Resolution */
        #     0   /* Accuracy */
        #     0   /* LowerLimit */
        #     4294967295   /* UpperLimit */
        #     READ_WRITE
        #     MATRIX_DIM 152 1 1
        #     ECU_ADDRESS 0x00000000
        # /end MEASUREMENT
        #
        # ----- Example blocks in a2l (Embedded Coder) -----
        #
        # /begin MEASUREMENT
        #   /* Name                   */      sVcAesVe_md_VolmcOffs
        #   /* Long identifier        */      "Volumetric cylinder mass flow offset"
        #   /* Data type              */      FLOAT32_IEEE
        #   /* Conversion method      */      VcAesVe_CM_Float32_g_s
        #   /* Resolution (Not used)  */      0
        #   /* Accuracy (Not used)    */      0
        #   /* Lower limit            */      -100.0
        #   /* Upper limit            */      100.0
        #   ECU_ADDRESS                       0x0000 /* @ECU_Address@sVcAesVe_md_VolmcOffs@ */
        # /end MEASUREMENT
        #
        # /begin CHARACTERISTIC
        #   /* Name                   */      cVcAesVe_D_VolmcCmpSel
        #   /* Long Identifier        */      "Select compensation factor characterizing deviation from nominal voleff"
        #   /* Type                   */      VALUE
        #   /* ECU Address            */      0x0000 /* @ECU_Address@cVcAesVe_D_VolmcCmpSel@ */
        #   /* Record Layout          */      Scalar_UBYTE
        #   /* Maximum Difference     */      0
        #   /* Conversion Method      */      VcAesVe_CM_uint8
        #   /* Lower Limit            */      1.0
        #   /* Upper Limit            */      3.0
        # /end CHARACTERISTIC

        self._block_finder = re.compile(r'(?:\s*\n)*'           # Optional blank lines
                                        r'(\s*/begin (\w+)\s*'  # begin <something> block
                                        r'\n\s*([\w.]+).*?\n'   # label. (Bosch-nvm contains the .)
                                        r'.*?'                  # block definition
                                        r'/end\s+\2)',          # end <something> block. Same something as before
                                        flags=re.M | re.DOTALL)

        self._tl_compu_method_parser = re.compile(
            r'(?:\s*\n)*(?P<compu_method>'
            r'\s*/begin COMPU_METHOD\s*\n'
            r'\s*(?P<name>\w*)\s*(/\* Name \*/)?\s*\n'              # Name
            r'\s*"(?P<ID>.*?)"\s*(/\* LongIdentifier \*/.*?)\s*\n'  # Long Identifier
            r'\s*(?P<conv_type>[A-Z_]*).*\s*\n'                     # ConversionType
            r'\s*"(?P<disp_format>.*)"\s*(/\* Format \*/)?\s*\n'    # Format
            r'\s*"(?P<unit>.*?)"\s*(/\* Unit \*/)?\s*\n'            # Unit
            r'\s*(?P<conversion>.*)\s*\n'                           # COEFFS
            r'(?P<indentation>\s*)/end COMPU_METHOD)', flags=re.M)  # No DOTALL, so .* is [^\n]*

        # COMPU_METHOD parser that works with files generated by Embedded Coder
        self._ec_compu_method_parser = re.compile(
            r'(?:\s*\n)*(?P<compu_method>'
            r'\s*/begin COMPU_METHOD\s*\n'
            r'\s*(/\* Name of CompuMethod\s*\*/)?\s*(?P<name>\w*)\s*\n'    # Name
            r'\s*(/\* Long identifier\s*\*/)?\s*"(?P<ID>.*?)"\s*\n'        # Long Identifier
            r'\s*/\* Conversion Type\s*\*/\s*(?P<conv_type>[A-Z_]*)\s*\n'  # ConversionType
            r'\s*(/\* Format\s*\*/)?\s*"(?P<disp_format>.*?)"\s*\n'        # Format
            r'\s*(/\* Units\s*\*/)?\s*"(?P<unit>.*?)"\s*\n'                # Unit
            r'\s*/\* Coefficients\s*\*/\s*(?P<conversion>.*?)\s*\n'        # COEFFS
            r'(?P<indentation>\s*)/end COMPU_METHOD)', flags=re.M)         # No DOTALL, so .* is [^\n]*

        self._expr_block_meas_kp_blob_parser = re.compile(
            r'\/begin\s+(?P<keyword>\w+)\s+(?P<class>\w+)\s*\n'
            r'\s*KP_BLOB\s+(?P<address>0x[0-9a-fA-F]+)\s*\n'
            r'\s*\/end\s+\1'
        )

        self._compu_methods = {}
        self._included_compu_methods = []

        self._tl_compu_method_template = Template(
            '$indentation/begin COMPU_METHOD\n'
            '$indentation    $name   /* Name */\n'
            '$indentation    "$ID"    /* LongIdentifier */\n'
            '$indentation    $conv_type    /* ConversionType */\n'
            '$indentation    "$disp_format"   /* Format */\n'
            '$indentation    "$unit" /* Unit */\n'
            '$indentation    $conversion\n'
            '$indentation/end COMPU_METHOD'
        )

        # COMPU_METHOD template that looks similar to COMPU_METHOD generated by Embedded Coder
        self._ec_compu_method_template = Template(
            '$indentation/begin COMPU_METHOD\n'
            '$indentation  /* Name of CompuMethod    */      $name\n'
            '$indentation  /* Long identifier        */      "$ID"\n'
            '$indentation  /* Conversion Type        */      $conv_type\n'
            '$indentation  /* Format                 */      "$disp_format"\n'
            '$indentation  /* Units                  */      "$unit"\n'
            '$indentation  /* Coefficients           */      $conversion\n'
            '$indentation/end COMPU_METHOD'
        )

        for filename in a2l_files_unit:
            removed_symbols = self._parse_unit(filename)
            self._removed_symbols.extend(removed_symbols)
            self.debug('Loaded %s', filename)
        for filename in a2l_files_gen:
            self._parse_gen(filename)
            self.debug('Loaded %s', filename)

    def _parse_unit(self, filename):
        """Parse the unit a2l-files and apply a filter to only active parameters."""
        self.debug('Processing %s', filename)
        with open(filename, 'r', encoding="ISO-8859-1") as a2lfp:
            a2ld = a2lfp.read()
            file_path_parts = os.path.split(filename)
            unit = file_path_parts[1].split('.')[0]
            base_path = file_path_parts[0]
            dcl_match = re.search(r'VcDcl[\w]+Mdl(__[\w]+)', base_path)
            if dcl_match is not None and 'Mdl' not in unit:
                # Hand coded model names including "__" will lead to this, due to name mismatch of .a2l and .json files.
                # E.g. VcDclPtrlMdl__denso:
                #     1. config_VcDclPtrlMdl__denso.json vs VcDclPtrlMdl__denso.a2l.
                #         1.1. Match: unit in self._per_unit_cfg.
                #     2. config_VcDclPtrl__denso.json vs VcDclPtrl.a2l.
                #         2.1. No match: unit not in self._per_unit_cfg.
                old_unit = unit
                unit = unit + dcl_match.group(1)
                self.info(
                    'Found unit %s with .a2l and .json file name mismatch. Using new unit name: %s',
                    old_unit,
                    unit
                )

            if unit in self._per_unit_cfg:
                u_conf = self._per_unit_cfg[unit]
                code_generator = u_conf['code_generator'] if 'code_generator' in u_conf else 'target_link'
            else:
                u_conf = {}
                code_generator = 'target_link'

            if code_generator == 'embedded_coder':
                blks = re.findall(r'(?:\s*\n)*(\s*/begin '
                                  r'(?!PROJECT|HEADER|MODULE|MOD_PAR|MOD_COMMON)(\w+)\s*(?:\n\s*)?'
                                  r'(?:/\*\s*[\w ]+\s*\*/\s*)?(\w+)([\[\d+\]]*).*?\n.*?/end\s+\2)',
                                  a2ld, flags=re.M | re.DOTALL)
            else:
                blks = re.findall(r'(?:\s*\n)*(\s*/begin (?!PROJECT|MODULE)(\w+)[\n\s]*'
                                  r'(\w+(?:\.\w+)?)([\[\d+\]]*).*?\n.*?/end\s+\2)', a2ld,
                                  flags=re.M | re.DOTALL)

            compu_method_translators = self._parse_compu_methods(a2ld, unit)
            unit_blks = {}
            removed_symbols = []
            if unit not in self._per_unit_cfg:
                # Handcoded a2l without json-files will lead to this.
                # Add json files for the handcoded a2l!
                # NOTE: Assuming TargetLink
                self.debug('%s is not in the units list. Looking for json.', unit)
                config_filename = os.path.join(
                    self._prj_cfg.get_unit_cfg_deliv_dir(),
                    f'config_{unit}.json')
                self.debug('Looking for %s', config_filename)
                if os.path.isfile(config_filename):
                    with open(config_filename, 'r', encoding="utf-8") as config_file:
                        u_conf = json.load(config_file)
                    self._handle_config(
                        code_generator, unit, u_conf, blks, unit_blks, removed_symbols, compu_method_translators
                    )
                else:
                    self.warning('%s does not have a unit_cfg json, '
                                 'including all a2l-parameters', unit)
                    for blk_def, type_, label, size in blks:
                        if type_ == 'COMPU_METHOD':
                            blk_def, label = self._replace_compu_method(blk_def, label, compu_method_translators)
                        self.add_block_definition(unit_blks, type_, label, size, blk_def, compu_method_translators)
            else:
                self._handle_config(
                    code_generator, unit, u_conf, blks, unit_blks, removed_symbols, compu_method_translators
                )
        deep_dict_update(self._blks, unit_blks)
        return removed_symbols

    def _handle_config(self, code_generator, unit, u_conf, blks, unit_blks, removed_symbols, compu_method_translators):
        """Merge all types of ram for the unit."""
        ram = u_conf['inports']
        ram.update(u_conf['outports'])
        ram.update(u_conf['local_vars'])
        # TODO: Function the variables and labels needs to be removed from
        # the FUNCTION block too
        for blk_def, type_, label, size in blks:
            remove_excluded_symbol = True
            inc = False
            if type_ == 'AXIS_PTS':
                if label in u_conf['calib_consts']:
                    inc = True
            elif type_ == 'CHARACTERISTIC':
                if label in u_conf['calib_consts']:
                    if label in [axis_label for _, axis_type, axis_label, _ in blks if axis_type == 'AXIS_PTS']:
                        # AXIS_PTS can be used as CHARACTERISTC but not the other way around.
                        # If there are duplicates, use the AXIS_PTS.
                        self.debug('Will not add the block for CHARACTERISTC %s, but will keep it as a symbol,'
                                   ' since it exists as AXIS_PTS', label)
                        remove_excluded_symbol = False
                        inc = False
                    else:
                        inc = self._handle_axis_ptr_ref_config(u_conf, blk_def, unit)
            elif type_ == 'MEASUREMENT':
                if label in ram:
                    key = label if size is None else label + size
                    if label in u_conf['outports']:
                        # This unit is producing the measurement.
                        inc = True
                    elif key in unit_blks.get(type_, {}):
                        # This unit is not producing it, and it has already been added
                        inc = False
                    else:
                        # This unit is not producing it, but it has not been added
                        # Could be external signal, etc.
                        inc = True
            elif type_ == 'COMPU_METHOD':
                inc = True
                blk_def, label = self._replace_compu_method(blk_def, label, compu_method_translators)
            else:
                inc = True
            if inc:
                self.add_block_definition(unit_blks, type_, label, size, blk_def, compu_method_translators)
            else:
                if remove_excluded_symbol:
                    removed_symbols.append(label + size)
                self.debug('Did not include A2L-blk %s%s', label, size)
                if not self._unit_cfg.check_if_in_unit_cfg(unit, label):
                    if type_ != 'COMPU_METHOD':
                        self.warning('A2l block %s not in config json file for %s', label, unit)

        if 'FUNCTION' in unit_blks:
            unit_blks['FUNCTION'] = self._remove_symbols_from_func_blks(
                code_generator, unit_blks['FUNCTION'], removed_symbols
            )

        if 'GROUP' in unit_blks:
            unit_blks['GROUP'] = self._remove_symbols_from_grp_blks(unit_blks['GROUP'], removed_symbols)

    def _handle_axis_ptr_ref_config(self, u_conf, blk, unit):
        """Remove blocks referencing undefined blocks."""
        ref_re = re.compile(r'\s*AXIS_PTS_REF\s*([\w]*)')
        for axis_ptr_ref in ref_re.findall(blk):
            if axis_ptr_ref not in u_conf['calib_consts']:
                self.debug('Excluding due to %s missing in config', axis_ptr_ref)
                return False
            if not self._unit_cfg.check_if_in_unit_cfg(unit, axis_ptr_ref):
                self.debug('Excluding due to %s not active in config', axis_ptr_ref)
                return False
        return True

    def add_block_definition(self, unit_blks, type_, label, size, blk_def, compu_method_translators):
        """Add block definition to A2L-file."""
        size = '' if size is None else size
        blk_def = self._replace_conversions(blk_def, compu_method_translators)
        if type_ not in unit_blks:
            unit_blks[type_] = {}
        unit_blks[type_][label + size] = blk_def

    @staticmethod
    def _parse_func_blk(code_generator, fnc_blk):
        """Remove the unused symbols from the FUNCTION blocks in the A2L-file.
        Parse the FUNCTION block, TL or EC style based on code_generator.
        """

        if code_generator == 'target_link':
            pattern = r'\s*/begin\s+FUNCTION\s*?\n\s*(\w+).*?\n\s*"(.*?)".*?\n(.*)'
        else:
            pattern = r'\s*/begin\s+FUNCTION\s*?\n\s*.*\*/\s*(\w+).*?\n\s*.*\*/\s*"(.*?)".*?\n(.*)'
        res = re.match(pattern, fnc_blk, flags=re.M | re.DOTALL)
        fnc_name = res.group(1)
        long_id = res.group(2)
        fnc_dict = {
            'fnc_name': fnc_name,
            'long_id': long_id,
            'body': {}
        }
        fnc_body = res.group(3)
        sb_res = re.findall(r'\s*/begin\s+(\w+[\[\d\]]*)\s*\n\s*'
                            r'(.*?\n)\s*/end \1', fnc_body, flags=re.M | re.DOTALL)
        for sb_name, sub_blk in sb_res:
            symbols = set(re.findall(r'\s*(\w+(?:\.\w+)?[\[\d\]]*).*?\n', sub_blk, flags=re.M))
            fnc_dict['body'][sb_name] = symbols
        return fnc_dict

    @staticmethod
    def _parse_grp_blk(grp_blk):
        """Remove the unused symbols from the GROUP blocks in the A2L-file."""
        # parse the GROUP block
        res = re.match(r'\s*/begin\s+GROUP\s*?\n\s*.*\*/\s*(\w+).*?\n\s*.*\*/\s*"(.*?)".*?\n(.*)',
                       grp_blk, flags=re.M | re.DOTALL)
        fnc_name = res.group(1)
        long_id = res.group(2)
        fnc_dict = {
            'fnc_name': fnc_name,
            'long_id': long_id,
            'body': {}
        }
        fnc_body = res.group(3)
        sb_res = re.findall(r'\s*/begin\s+(\w+[\[\d\]]*)\s*\n\s*'
                            r'(.*?\n)\s*/end \1', fnc_body, flags=re.M | re.DOTALL)
        for sb_name, sub_blk in sb_res:
            symbols = set(re.findall(r'\s*(\w+(?:\.\w+)?[\[\d\]]*).*?\n', sub_blk, flags=re.M))
            fnc_dict['body'][sb_name] = symbols
        return fnc_dict

    def _recursive_remove(self, a2l_dict, name):
        """Remove symbols from A2L dict (e.g. group or function)."""
        if name in a2l_dict:
            blk = a2l_dict[name]
            if 'SUB_FUNCTION' in blk:
                for sub_fnc in blk['SUB_FUNCTION']:
                    if self._recursive_remove(a2l_dict, sub_fnc):
                        blk['SUB_FUNCTION'] = blk['SUB_FUNCTION'] - set([sub_fnc])
            elif 'SUB_GROUP' in blk:
                for sub_grp in blk['SUB_GROUP']:
                    if self._recursive_remove(a2l_dict, sub_grp):
                        blk['SUB_GROUP'] = blk['SUB_GROUP'] - set([sub_grp])
            empty = True
            for key in blk:
                if blk[key]:
                    empty = False
                    break
            if empty:
                a2l_dict.pop(name)
                return True
        return False

    def _remove_symbols_from_func_blks(self, code_generator, fnc_blks, removed_symbols):
        """Remove the unused symbols from function blocks.

        If the function block is empty, it too will be removed.
        first iteration - remove all symbols that have been removed
        second iteration - recusively remover all functions without symbols
        """
        fnc_dict = {}
        for fnc_name, fnc_blk in fnc_blks.items():
            fnc_dict[fnc_name] = {}
            u_fnc_bdy = self._parse_func_blk(code_generator, fnc_blk)['body']
            sub_blk_types = set(u_fnc_bdy.keys()) - set(['SUB_FUNCTION'])
            for type_ in list(sub_blk_types):
                fnc_dict[fnc_name][type_] = u_fnc_bdy[type_] - set(removed_symbols)
            if 'SUB_FUNCTION' in u_fnc_bdy:
                fnc_dict[fnc_name]['SUB_FUNCTION'] = u_fnc_bdy['SUB_FUNCTION']
        # second iteration - remove empty FUNCTION blocks
        # TODO: Add functionality which parses the the function tree structures
        # And the run recursive remove on all tree roots.
        for fnc_name in fnc_blks.keys():
            self._recursive_remove(fnc_dict, fnc_name)
        # generate new function blocks
        new_fnc_blks = {}
        for fnc_name, fnc_data in fnc_dict.items():
            fnc_blk = f'    /begin FUNCTION\n        {fnc_name}\t/* Name */\n'
            fnc_blk += "        \"\"\t/* LongIdentifier */\n"
            for sub_sec in sorted(fnc_data.keys()):
                sub_sec_data = fnc_data[sub_sec]
                if sub_sec_data:
                    fnc_blk += f"        /begin {sub_sec}\n"
                    for param in sorted(sub_sec_data):
                        fnc_blk += f"            {param}\t/* Identifier */\n"
                    fnc_blk += f"        /end {sub_sec}\n"
            fnc_blk += "    /end FUNCTION"
            new_fnc_blks[fnc_name] = fnc_blk
        return new_fnc_blks

    def _remove_symbols_from_grp_blks(self, grp_blks, removed_symbols):
        """Remove the unused symbols from group blocks.

        If the group block is empty, it too will be removed.
        first iteration - remove all symbols that have been removed
        second iteration - recusively remover all groups without symbols
        """
        grp_dict = {}
        for grp_name, grp_blk in grp_blks.items():
            grp_dict[grp_name] = {}
            u_grp_bdy = self._parse_grp_blk(grp_blk)['body']
            sub_blk_types = set(u_grp_bdy.keys()) - set(['SUB_GROUP'])
            for type_ in list(sub_blk_types):
                grp_dict[grp_name][type_] = u_grp_bdy[type_] - set(removed_symbols)
            if 'SUB_GROUP' in u_grp_bdy:
                grp_dict[grp_name]['SUB_GROUP'] = u_grp_bdy['SUB_GROUP']
        # second iteration - remove empty GROUP blocks
        # TODO: Add functionality which parses the the group tree structures
        # And the run recursive remove on all tree roots.
        for grp_name in grp_blks.keys():
            self._recursive_remove(grp_dict, grp_name)
        # generate new group blocks
        new_grp_blks = {}
        for grp_name, grp_data in grp_dict.items():
            grp_blk = f"    /begin GROUP \n      /* Name                   */       {grp_name}\n"
            grp_blk += "      /* Long identifier        */       \"\"\n"
            for sub_sec in sorted(grp_data.keys()):
                sub_sec_data = grp_data[sub_sec]
                if sub_sec_data:
                    grp_blk += f"      /begin {sub_sec}\n"
                    for param in sorted(sub_sec_data):
                        grp_blk += f"        {param}\n"
                    grp_blk += f"      /end {sub_sec}\n"
            grp_blk += "    /end GROUP"
            new_grp_blks[grp_name] = grp_blk

        return new_grp_blks

    def _parse_gen(self, filename):
        """Parse the generated a2l-files, without filter."""
        self.debug('parsing gen a2l: %s', filename)
        with open(filename, 'r', encoding="utf-8") as a2lfp:
            a2ld = a2lfp.read()
            for blk_def, type_, label in self._block_finder.findall(a2ld):
                self._blks.setdefault(type_, {}).setdefault(label, blk_def)

    @staticmethod
    def _replace_compu_method(blk_def, label, compu_method_translators):
        """Replace the compu method block and label."""
        for translator in compu_method_translators:
            if translator['old_compu_method'] == blk_def:
                return translator['new_compu_method'], translator['new_name']
        return blk_def, label

    def _store_compu_method(self, ID, conv_type, disp_format, unit, conversion, indentation, u_conf):
        """Stash compu methods that exists in the resulting a2l."""
        key = (ID, conv_type, disp_format, unit, conversion)
        if key in self._compu_methods:
            new_name = self._compu_methods[key]['name']
            new_compu_method = self._compu_methods[key]['method']
        else:
            new_name = 'Scaling_' + str(len(self._compu_methods))
            if 'code_generator' in u_conf and u_conf['code_generator'] == 'embedded_coder':
                new_compu_method = self._ec_compu_method_template.substitute(
                    name=new_name,
                    ID=ID,
                    conv_type=conv_type,
                    disp_format=disp_format,
                    unit=unit,
                    conversion=conversion,
                    indentation=indentation
                )
            else:
                new_compu_method = self._tl_compu_method_template.substitute(
                    name=new_name,
                    ID=ID,
                    conv_type=conv_type,
                    disp_format=disp_format,
                    unit=unit,
                    conversion=conversion,
                    indentation=indentation
                )
            self._compu_methods.update({key: {'name': new_name,
                                              'method': new_compu_method}})
        return new_name, new_compu_method

    @staticmethod
    def _replace_conversions(blk_def, compu_method_translators):
        """Replace conversion identifiers in a2l block."""
        for translator in compu_method_translators:
            # The following check is faster than running the regex on the block.
            # It DOES give false positives, which is why the regex is used for substitution
            # and we do not immidiately return after one positive
            if translator['old_name'] in blk_def:
                blk_def = translator['regex'].sub(translator['replacement'], blk_def)
        return blk_def

    def _parse_compu_methods(self, blk_def, unit):
        """Replace compu methods to not overwrite any of them."""
        compu_method_translators = []  # Translators for one processed a2l file. Needs to be reset between files
        u_conf = self._per_unit_cfg[unit] if unit in self._per_unit_cfg else {}
        if 'code_generator' in u_conf and u_conf['code_generator'] == 'embedded_coder':
            for match in self._ec_compu_method_parser.finditer(blk_def):
                new_name, new_compu_method = self._store_compu_method(
                    match['ID'],
                    match['conv_type'],
                    match['disp_format'],
                    match['unit'],
                    match['conversion'],
                    match['indentation'],
                    u_conf
                )
                compu_method_translators.append(
                    {
                        'new_name': new_name,
                        'old_name': match['name'],
                        'regex': re.compile(
                            r'(\s*)'                                     # beginning
                            r'\s*(/\* Conversion [Mm]ethod\s*\*/\s*)'    # optional comment
                            r'\b{name}\b'                                # word
                            r'('
                            r'\s*\n'                                     # newline
                            r')'.format(name=match['name'])              # end of end-match
                        ),
                        'replacement': r'\1\2{name}\3'.format(name=new_name),
                        'old_compu_method': match['compu_method'],
                        'new_compu_method': new_compu_method
                    }
                )
        else:
            for match in self._tl_compu_method_parser.finditer(blk_def):
                new_name, new_compu_method = self._store_compu_method(
                    match['ID'],
                    match['conv_type'],
                    match['disp_format'],
                    match['unit'],
                    match['conversion'],
                    match['indentation'],
                    u_conf
                )
                compu_method_translators.append(
                    {
                        'new_name': new_name,
                        'old_name': match['name'],
                        'regex': re.compile(
                            r'(\s*)'                         # beginning
                            r'\b{name}\b'                    # word
                            r'('                             # start of end-match
                            r'\s*(/\* Conversion \*/)?'      # optional comment
                            r'\s*\n'                         # newline
                            r')'.format(name=match['name'])  # end of end-match
                        ),
                        'replacement': r'\1{name}\2'.format(name=new_name),
                        'old_compu_method': match['compu_method'],
                        'new_compu_method': new_compu_method
                    }
                )
        return compu_method_translators

    def _patch_kp_blob(self, block):
        """Return updated measurement block text.
        Args:
            block (str): A2L text block
        Returns:
            a2l_text (str): A2L text without KP_BLOB.
        """
        ecu_address = '0x00000000'
        for match in self._expr_block_meas_kp_blob_parser.finditer(block):
            start, end = match.span()
            block = f'{block[:start]}ECU_ADDRESS {ecu_address}{block[end:]}'
        return block

    def merge(self, f_name, complete_a2l=False, silver_a2l=False):
        """Write merged a2l-file.

        Args:
            f_name (str): Output filename.
        """
        a2l = ''
        a2l_config = self._prj_cfg.get_a2l_cfg()
        for _, data in self._blks.items():
            for _, blk in data.items():
                if not a2l_config['allow_kp_blob']:
                    blk = self._patch_kp_blob(blk)
                a2l += blk + '\n\n'

        if complete_a2l:
            events = []
            time_unit_10ms = '0x07'
            rasters = self._prj_cfg.get_units_raster_cfg()
            for xcp_id, evt_data in enumerate(rasters['SampleTimes'].items(), 1):
                events.append({
                    'time_cycle': '0x%02X' % int(evt_data[1] * 100),
                    'time_unit': time_unit_10ms,
                    'name': evt_data[0],
                    'channel_id': '0x%04X' % xcp_id
                })
            a2l_template = A2lProjectTemplate(
                a2l,
                a2l_config['asap2_version'],
                a2l_config['name'],
                events,
                a2l_config['ip_address'],
                a2l_config['ip_port']
            )
            a2l = a2l_template.render()
        elif silver_a2l:
            a2l_template = A2lSilverTemplate(a2l)
            a2l = a2l_template.render()

        self.a2l = a2l

        with open(f_name, 'w', encoding="ISO-8859-1") as ma2l:
            ma2l.write(a2l)
            self.info('Written the merged A2L-file %s', f_name)

    def get_characteristic_axis_data(self):
        """Get characteristic map axis data from merged a2l-file."""
        axis_data = {}
        for blk_def, type_, label in self._block_finder.findall(self.a2l):
            if type_ == "CHARACTERISTIC":
                axes = re.findall('AXIS_PTS_REF (.*)', blk_def)
                if label in axis_data:
                    self.critical("Multiple CHARACTERISTIC for %s in merged a2l.", label)
                axis_data[label] = {'axes': axes}
        return axis_data



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Library for using matlab.

Called pt_matlab to not collide with Mathwork's matlab package.
"""

import logging
import os
import platform
import subprocess
import sys
import time

my_system = platform.system()
if my_system == "Darwin":
    MATLAB_BIN = os.environ.get(
        "MatInstl",
        os.environ.get("MatInstl2017", "/Applications/MATLAB_R2020a.app/bin/matlab")
    )
elif my_system == "Linux":
    MATLAB_BIN = os.environ.get("MatInstl", os.environ.get("MatInstl2017", "/usr/local/MATLAB/R2017b/bin/matlab"))
else:
    from . import pt_win32
    MATLAB_BIN = os.environ.get("MatInstl", os.environ.get("MatInstl2017", "C:\\MATLABR2017b_x64\\bin\\matlab.exe"))

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
RETRY_TIME = 10
POLL_TIME = 1


def get_env():
    """Get environment variables."""
    env = os.environ.copy()
    workspace = os.environ.get("WORKSPACE", "")
    if workspace:
        env["MATLAB_PREFDIR"] = os.path.join(workspace, "ConfigDocuments", "buildSlaveMatlabSettings")
    return env


def wrap_command(command):
    """Wrap a command in a try-catch with an exit at the end.

    Use this to be sure never to raise an error in matlab that causes a hanging matlab.
    This is useful when writing matlab scripts to be used in jenkins jobs.

    Args:
        command (str): command to wrap in try-catch-exit.
    Returns:
        command (str): the command wrapped in try-catch-exit
    """
    return f"try;exitcode=0;{command};catch err;disp(getReport(err));exitcode=1;end;exit(exitcode);"


def wrap_m_script(path_to_mscript):
    """Wrap a path to an m-script to ensure no jobs in jenkins are left hanging.

    This is useful when writing matlab scripts to be used in jenkins jobs.

    Args:
        path_to_mscript (str): file to run in matlab.
    Returns:
        command (str): the command to send to matlab to run the script
    """
    mscript_folder, mscript_file = os.path.split(path_to_mscript)
    LOGGER.debug("Using folder %s for the m-script", mscript_folder)
    if mscript_folder:
        command = wrap_command(
            f"addpath('{mscript_folder}');assert(boolean(exist('{mscript_file[:-2]}')));{mscript_file[:-2]}"
        )
    else:
        command = wrap_command(f"assert(boolean(exist('{mscript_file[:-2]}')));{mscript_file[:-2]}")
    return f'"{command}"'


def list_to_cell(python_list):
    """Convert a python list to a cell argument string for matlab.

    Args:
        python_list (list): A list in python
    Returns:
        cell (str): A cell to be read by Matlab.
    """
    matlab_list = "', '".join(python_list)
    return f"{{'{matlab_list}'}}"


def write_m_script(mfile, base_cmd, wrap_cmd=True):
    """Write an m-script to run from the command line.

    The m-script will exit Matlab with an exitcode.
    This exitcode can be set in the base_cmd if needed.

    Args:
        mfile (str): path to m-file to create
        base_cmd (str): Base command to run in Matlab
        wrap_cmd (bool): wrap command
    """
    if wrap_cmd:
        cmd = wrap_command(base_cmd)
    else:
        cmd = base_cmd
    LOGGER.debug("Writing %s", mfile)
    with open(mfile, mode="w", encoding="utf-8") as m_file:
        m_file.writelines(cmd.split(os.linesep))


def is_model(model_file):
    """Check if a file name is a simulink model or not.

    Args:
        model_file (str): filename to check
    """
    return model_file.endswith(".mdl") or model_file.endswith(".slx")


def extract_model_names(files, ending):
    """Find model names from path to files.

    Checks if the files corresponds to model files.

    Args:
        files (list): files that potentially corresponds to models
        ending (str): file ending that is used for models.
    Returns:
        found_models (list): Models to run checkscripts on.
    """
    found_models = []
    for found_file in files:
        name = os.path.basename(found_file)
        if name.endswith(ending):
            name = name[: -len(ending)]
            found_models.append(name)
    return found_models


def cmd_path(path, subdirs=False):
    """Generate Matlab command text for adding supplied path and subdirectories.

    Args:
        path (str): Path to generate commands for adding.
        subdirs (bool): If True, recursively adds subdirs. Default: False

    Returns:
        cmds (list): Commands as a list of strings.
    """
    return f"addpath(genpath('{path}'))" if subdirs else f"addpath('{path}')"


def cmd_callfunc(func, *args):
    """Generate Matlab command for calling a function with optional arguments.

    Example: cmd_callfunc('my_function')
             cmd_callfunc('my_function', 'argument 1', 'argument 2')

    Args:
        func (str): Name of function to call.
        args (argument list): Variadic argument list to pass to function.
    Returns:
        cmd (str): Command text for calling function.
    """
    args = ", ".join([cmd_arg(arg) for arg in args])
    return f"{func}({args})"


def cmd_arg(arg):
    """Transform Python argument to Matlab argument."""
    if isinstance(arg, str):
        return f"'{arg}'"
    if isinstance(arg, bool):
        return "true" if arg else "false"
    if isinstance(arg, list):
        return list_to_cell(arg)
    raise ValueError("Unable to convert argument to Matlab type")


def cmds_join(cmds):
    """Join several Matlab command strings together, separated by terminators and newlines.

    Args:
        cmds (list): List of command string to join together.
    Returns:
        cmds (str): Command text for executing all supplied commands.
    """
    return f";{os.linesep}".join(cmds)


class TargetlinkWatcher(logging.StreamHandler):
    """Class to check output log for known error messages."""

    intermittent_errors = [
        "Undefined function 'tl_pref' for input arguments of type 'char'",
        "Undefined function 'dsdd_manage_project' for input arguments of type 'char'.",
        "Exception: ACCESS_VIOLATION",
    ]

    def __init__(self, stream):
        """Init."""
        super().__init__(stream)
        self.targetlink_error = False

    def reset_errors(self):
        """Reset errors."""
        self.targetlink_error = False

    def emit(self, record):
        """Set error flag on error message in log."""
        for intermittent_error in self.intermittent_errors:
            if intermittent_error in record.getMessage():
                self.targetlink_error = True
        super().emit(record)


class MatlabWatcher(logging.StreamHandler):
    """Class to check output log for sucess message."""

    success_message = "Matlab task succeeded!"

    def __init__(self):
        """Init."""
        super().__init__()
        self.task_success = False

    def reset_errors(self):
        """Reset errors."""
        self.task_success = False

    def emit(self, record):
        """Set success flag on success message in log."""
        if self.success_message in record.getMessage():
            self.task_success = True


class Matlab:
    """Wrapper for calling matlab from python scripts.

    The wrapper runs a m-script and optionally pipes the logs to stdout.
    """

    def __init__(self, log=None, dry_run=False, matlab_bin=MATLAB_BIN, nojvm=True):
        """Init.

        Args:
            log (str): file name to log to. Default: None
            dry_run (bool): Run a mocked version of matlab. Default: False
            matlab_bin (str): path to matlab binary to use in execution
            nojvm (bool): False if jvm should be used. Default: True
        """
        self.log = log
        self.dry_run = dry_run
        self.nojvm = nojvm
        self.targetlink_watcher = TargetlinkWatcher(sys.stdout)
        self.matlab_watcher = MatlabWatcher()
        self.mlogger = self.create_logger()
        self.mlogger.addHandler(self.matlab_watcher)
        self.open_log_file = None
        self.matlab_bin = matlab_bin

    @staticmethod
    def add_args(parser):
        """Parse arguments."""
        matlab_parser = parser.add_argument_group("matlab arguments")
        matlab_parser.add_argument(
            "--matlab-bin",
            help=f"Path to the matlab binary to use. Defaults to {MATLAB_BIN}.",
            default=MATLAB_BIN,
        )

    def create_logger(self):
        """Create logger for matlab.

        Returns:
            logger (logging.Logger): Logger for matlab.
        """
        handler = self.targetlink_watcher
        handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(message)s"))
        logger = logging.getLogger("matlab")
        logger.addHandler(handler)
        return logger

    def sync_log(self, last_stat):
        """Read the latest loglines from the matlab log and write to log.

        Checks if the matlab log file has been updated since last call and
        redirects new entries to the python log.
        If self.log is None, this function returns None at once.

        Args:
            last_stat (os.stat): last timestamp for an updated log.
        Returns:
            new_stat (os.stat): new timestamp for an updated log.
        """
        if self.open_log_file is None:
            return None
        current_position = self.open_log_file.tell()
        if os.stat(self.open_log_file.name) == last_stat:
            return last_stat  # Do not read partial lines again.
        for line in self.open_log_file.readlines():
            # The non-finished lines behaviour was sometimes seen when using diary.
            # It should not hurt to have, but if it is causing problems, just remove it.
            # It is purely cosmetic.
            if line.endswith("\n"):
                if line != "\n":  # Skip the end of the line we jumped to from a non-finished line
                    self.mlogger.info(line[0:-1])
            else:
                LOGGER.debug("Partial line read. Go back and wait for update.")
                self.open_log_file.seek(current_position - len(line))
                break
            current_position = self.open_log_file.tell()
        return os.stat(self.open_log_file.name)

    def compose_m_script_cmd(self, mfile, wrap_cmd):
        """Compose batch script to run matlab."""

        if my_system in {"Darwin", "Linux"}:
            cmd = [
                self.matlab_bin if not self.dry_run else "echo",
                "-nodisplay",
                "-nodesktop",
            ]
        else:
            cmd = [
                self.matlab_bin if not self.dry_run else "cmd rem",
                "-nodisplay",
                "-nodesktop",
                "-wait",
                "-minimize",
            ]

        if self.log:
            cmd += ["-logfile", self.log]
        if self.nojvm:
            cmd += ["-nojvm"]
        if wrap_cmd:
            if my_system == "Darwin":
                cmd += ["<", wrap_m_script(mfile)]
            else:
                cmd += ["-r", wrap_m_script(mfile)]
        else:
            if my_system == "Darwin":
                cmd += ["<", mfile]
            else:  # Linux will be similar to Windows
                cmd += ["-r", mfile[:-2]]

        return cmd

    def run_m_script(self, mfile, wrap_cmd=True, attempts=10):
        """Run the composed m-script.

        Args:
            mfile (str): The composed m-script
            wrap_cmd (bool): wrap command (Default: True)
            attempts (int): Number of times to run the command again if an intermittent problem
                            is detected (Default: 1)
        Returns:
            process_status: Exitcode from Matlab (or 1 if Matlab was never started)
        """

        def wait_for_process(p_matlab, stat=None):
            stat = self.sync_log(None)  # Initial sync in case matlab has already finished

            if my_system in {"Darwin", "Linux"}:
                while p_matlab.poll() is None:
                    time.sleep(POLL_TIME)
                    stat = self.sync_log(stat)
            else:
                while p_matlab.poll() == pt_win32.STILL_ACTIVE:
                    time.sleep(POLL_TIME)
                    stat = self.sync_log(stat)

        if attempts <= 0:
            LOGGER.error("Already attempted this too many times. Exiting.")
            return 2  # Do not trust previous exit codes

        if self.log and os.path.isfile(self.log):
            os.remove(self.log)

        p_matlab = None
        process_status = 1  # Something could go wrong in the try, before the exitcode is set

        try:
            env = get_env()
            LOGGER.debug("Running in dry_run mode: %s", self.dry_run)
            cmd = self.compose_m_script_cmd(mfile, wrap_cmd)
            LOGGER.debug("Calling: %s", " ".join(cmd))

            if my_system in {"Darwin", "Linux"}:
                p_matlab = subprocess.Popen(cmd)
            else:
                p_matlab = pt_win32.PtWin32Process()
                p_matlab.set_environment(env)
                p_matlab.run(command=" ".join(cmd))

            if not self.dry_run:
                if self.log:
                    # Wait for log file to be created
                    if my_system in {"Darwin", "Linux"}:
                        while not os.path.isfile(self.log) and p_matlab.poll() is None:
                            time.sleep(POLL_TIME)
                    else:
                        while not os.path.isfile(self.log) and p_matlab.poll() == pt_win32.STILL_ACTIVE:
                            time.sleep(POLL_TIME)

                self.open_log_file = open(self.log, "r", encoding="utf-8", errors='ignore') if self.log else None

            wait_for_process(p_matlab)
            process_status = p_matlab.poll()

            LOGGER.info("Matlab returned with exitcode %s", process_status)
            if self.targetlink_watcher.targetlink_error:
                LOGGER.warning("Found an intermittent targetlink problem. Trying again, attempts left: %s", attempts)
                if self.open_log_file:
                    self.open_log_file.close()
                self.targetlink_watcher.reset_errors()
                time.sleep(RETRY_TIME)
                process_status = self.run_m_script(mfile, wrap_cmd=wrap_cmd, attempts=attempts - 1)
        finally:
            if my_system in {"Darwin", "Linux"}:
                if p_matlab.poll() is not None:
                    p_matlab.terminate()
                    p_matlab.wait()
            else:
                if p_matlab.poll() == pt_win32.STILL_ACTIVE:
                    p_matlab.terminate()

            if self.open_log_file:
                self.open_log_file.close()
        return process_status



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Feature configuration (codeswitches) module."""

import copy
import glob
import os
import re
from pprint import pformat

from powertrain_build.lib.helper_functions import deep_dict_update
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.xlrd_csv import WorkBook


class FeatureConfigs(ProblemLogger):
    """Hold feature configurations read from SPM_Codeswitch_Setup*.csv config files.

    Provides methods for retrieving the currently
    used configurations of a unit.
    """

    convs = (('~=', '!='), ('~', ' not '), ('!', ' not '), (r'\&\&', ' and '),
             (r'\|\|', ' or '))

    def __init__(self, prj_config):
        """Constructor.

        Args:
            prj_config (BuildProjConfig): configures which units are active in the current project and where
                                          the codeswitches files are located

        """
        super().__init__()
        self._if_define_dict = {}
        self._build_prj_config = prj_config
        self._missing_codesw = set()
        # Get the config switches configuration
        self._set_config(self._parse_all_code_sw_configs())
        self._parse_all_local_defs()
        self._add_local_defs_to_tot_code_sw()

    def __repr__(self):
        """Get string representation of object."""
        return pformat(self.__code_sw_cfg.keys())

    def _parse_all_code_sw_configs(self):
        """Parse all SPM_Codeswitch_Setup*.csv config files.

        Returns:
            dict: with the projects as keys, and the values are
                another dict with the config-parameter and it's value.

        """
        # TODO: Change this when condeswitches are moved to model config
        # cfg_paths = self._build_prj_config.get_unit_mdl_dirs('all')
        cfg_paths = [self._build_prj_config.get_prj_cfg_dir()]
        cfg_fname = self._build_prj_config.get_codeswitches_name()
        cfg_files = []
        for cfg_path in cfg_paths:
            cfg_files.extend(glob.glob(os.path.join(cfg_path, cfg_fname)))
        self.debug('cfg_paths: %s', pformat(cfg_paths))
        self.debug('cfg_fname: %s', pformat(cfg_fname))
        self.debug('cfg_files: %s', pformat(cfg_files))
        conf_dict = {}
        for file_ in cfg_files:
            conf_dict = deep_dict_update(conf_dict, self._parse_code_sw_config(file_))
        return conf_dict

    def _parse_code_sw_config(self, file_name):
        """Parse the SPM_Codeswitch_Setup.csv config file.

        Returns:
            dict: with the projects as keys, and the values are
                another dict with the config-parameter and it's value.

        """
        self.debug('_parse_code_sw_config: %s', file_name)
        wbook = WorkBook(file_name)
        conf_dict = {'NEVER_ACTIVE': 0,
                     'ALWAYS_ACTIVE': 1}
        # TODO: handle sheet names in a better way!
        wsheet = wbook.single_sheet()
        prjs = [d.value for d in wsheet.row(0)[2:]]
        prj_row = enumerate(prjs, 2)
        for col, prj in prj_row:
            if prj != self._build_prj_config.get_prj_config():
                self.debug('Skipping %s', prj)
                continue
            for r_nbr in range(1, wsheet.nrows):
                row = wsheet.row(r_nbr)
                conf_par = row[0].value.strip().replace('.', '_')
                val = row[col].value
                if not isinstance(val, str):
                    conf_dict[conf_par] = val
                elif val.lower().strip() == 'na' or val.lower().strip() == 'n/a':
                    conf_dict[conf_par] = 0
                else:
                    self.warning('Unexpected codeswitch value %s = "%s". Ignored!', row[0].value.strip(), val)
            return conf_dict
        return conf_dict

    def _recursive_subs(self, m_def, code_sws):
        """Recursivly replaces macro definitions with values."""
        # find and replace all symbols with values
        symbols = re.findall(r'(?!(?:and|or|not)\b)(\b[a-zA-Z_]\w+)', m_def)
        m_def_subs = m_def
        for symbol in symbols:
            if symbol in code_sws:
                m_def_subs = re.sub(symbol, str(code_sws[symbol]), m_def_subs)
            elif symbol in self._if_define_dict:
                m_def_subs = re.sub(symbol, str(self._if_define_dict[symbol]), m_def_subs)
                m_def_subs = self._recursive_subs(m_def_subs, code_sws)
            else:
                self.critical('Symbol %s not defined in config switches.', symbol)
                return None
        return m_def_subs

    def _add_local_defs_to_tot_code_sw(self):
        """Add the defines from the LocalDefs.h files to the code switch dict."""
        for macro, m_def in self._if_define_dict.items():
            tmp_subs = self._recursive_subs(m_def, self.__tot_code_sw)
            if tmp_subs is None:
                continue
            self.__tot_code_sw[macro] = eval(tmp_subs)

    def get_preprocessor_macro(self, nested_code_switches):
        """Get the #if macro string for a code switch configuration from a unit config json file.

        Args:
            nested_code_switches(list()): list of lists of code switches from unitconfig
        return:
            string: A string with an #if macro that defines if the code should be active
                    '#if (<CS1> && <CS2>) || (<CS3> && <CS4>)'
        """
        if_macro_and = []
        if not isinstance(nested_code_switches, list):
            self.warning("Unitconfig codeswitches should be in a nested list")
            nested_code_switches = [nested_code_switches]
        if not isinstance(nested_code_switches[0], list):
            self.warning("Unitconfig codeswitches should be in a nested list")
            nested_code_switches = [nested_code_switches]

        for code_switches in nested_code_switches:
            if isinstance(code_switches, str):
                code_switches = [code_switches]
            if_macro_and.append(f"( { ' && '.join(code_switches) } )")
        if_macro_string = f"#if {' || '.join(if_macro_and)}" if if_macro_and else ""
        all_projects = re.search('all', if_macro_string, re.I)
        if all_projects:
            return ""
        return if_macro_string

    def gen_unit_cfg_header_file(self, file_name):
        """Generate a header file with preprocessor defines needed to configure the SW.

        Args:
            file_name (str): The file name (with path) of the unit config header file

        """
        with open(file_name, 'w', encoding="utf-8") as f_hndl:
            _, fname = os.path.split(file_name)
            fname = fname.replace('.', '_').upper()
            f_hndl.write(f'#ifndef {fname}\n')
            f_hndl.write(f'#define {fname}\n\n')
            conf_sw = self.__code_sw_cfg
            for key_, val in conf_sw.items():
                if val == "":
                    self.warning('Code switch "%s" is missing a defined value', key_)
                f_hndl.write(f'#define {key_} {val}\n')
            f_hndl.write(f'\n#endif /* {fname} */\n')

    def _eval_cfg_expr(self, elem):
        """Convert matlab config expression to python expression.

        Uses the tuple self.convs, and evaluates the result using
        self.__code_sw_cfg[config]
        This function does not handle the complex definitions made outside
        the dict.

        Args:
            elem (str): element string

        Returns:
            Bool: True if config is active, False if not.

        """
        res = re.search('all', elem, re.I)
        if res is not None:
            return True
        # modify all matlab expressions
        elem_tmp = elem
        # find and replace all symbols with values
        symbols = re.findall(r'[a-zA-Z_]\w+', elem_tmp)
        code_sw_dict = self.__tot_code_sw
        for symbol in symbols:
            try:
                elem_tmp = re.sub(fr'\b{symbol}\b', str(code_sw_dict[symbol]), elem_tmp)
            except KeyError:
                if symbol not in self._missing_codesw:
                    self.critical('Missing %s in CodeSwitch definition', symbol)
                    self._missing_codesw.add(symbol)
                return False
        # convert matlab/c to python expressions
        for conv in self.convs:
            elem_tmp = re.sub(conv[0], conv[1], elem_tmp)
        # evaluate and return result
        return eval(elem_tmp)

    def check_if_active_in_config(self, config_def):
        """Check if a config is active in the current context.

        Takes a collection of config strings and checks if this config definition is active
        within the current configuration. The structure of the provided string collection
        determines how the config definition is evaluated. (logical and/or expressions)

        list of list of config strings
            [[*cs1* and *cs2*] or [*cs3* and *cs4*]].

        list of config strings
            [*cs1* and *cs2*]

        single config string
            *cs1*

        Args:
            config_def (list): the config definitions as described above

        Returns:
            Bool: True if active the current configuration

        """
        if not config_def:
            return True
        # format the input to a list of list of strings
        if isinstance(config_def, str):
            c_def = [[config_def]]
        elif isinstance(config_def, list) and isinstance(config_def[0], str):
            c_def = [config_def]
        else:
            c_def = config_def
        eval_ = False
        for or_elem in c_def:
            for and_elem in or_elem:
                eval_ = self._eval_cfg_expr(and_elem)
                if not eval_:
                    break
            if eval_:
                break
        return eval_

    def _conv_mlab_def_to_py(self, matlab_def):
        """Convert matlab syntax to python syntax.

        TODO: Move this functionality to the matlab-scripts, which are
        run on the local machine.

        """
        m_def_tmp = matlab_def
        for from_, to_ in self.convs:
            m_def_tmp = re.sub(from_, to_, m_def_tmp)
        return m_def_tmp

    def _parse_local_def(self, file_data):
        """Parse one local define file."""
        res = re.findall(r'#if\s+(.*?)(?<!\\)$\s*#define (\w+)\s+#endif',
                         file_data, flags=re.M | re.DOTALL)
        def_wo_if_endif = re.sub(r'#if(.*?)#endif', '', file_data, flags=re.M | re.DOTALL)
        one_line_def = re.findall(r'#define\s+(\w+)\s+(.+?)(?:(?://|/\*).*?)?$',
                                  def_wo_if_endif, flags=re.M)
        res.extend([(v, d) for (d, v) in one_line_def])
        # remove line break '\' characters and line break
        self._if_define_dict.update({d: self._conv_mlab_def_to_py(re.sub(r'\s*?\\$\s*', ' ', i))
                                     for (i, d) in res})

    def _parse_all_local_defs(self):
        """Parse all local define files."""
        def_paths = self._build_prj_config.get_unit_src_dirs()
        ld_fname = self._build_prj_config.get_local_defs_name()
        loc_def_files = []
        for def_path in def_paths.values():
            loc_def_files.extend(glob.glob(os.path.join(def_path, ld_fname)))
        for file_ in loc_def_files:
            with open(file_, 'r', encoding="utf-8") as fhndl:
                data = fhndl.read()
                self._parse_local_def(data)
        self.debug('self._if_define_list:\n%s\n', pformat(self._if_define_dict))

    def _set_config(self, code_sw):
        """Set config for code switches.

        Useful for unit testing.

        Args:
            code_sw
        """
        # __tot_code_sw
        self.__code_sw_cfg = code_sw
        self.__tot_code_sw = copy.deepcopy(self.__code_sw_cfg)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for calculating interfaces for CSP"""
import argparse
import sys
from pathlib import Path
from typing import List, Optional

from powertrain_build.interface.hal import HALA, get_hal_list
from powertrain_build.interface.device_proxy import DPAL
from powertrain_build.interface.service import ServiceFramework, get_service_list
from powertrain_build.lib import logger
from powertrain_build.interface import generation_utils

LOGGER = logger.create_logger("CSP wrappers")

PARSER_HELP = "Generate specifications for pt-scheduler wrappers"


def get_manifest(app, domain, client_name):
    """Get signal manifest for application

    The model yaml files are parsed for the models included in the application.
    If there are no signals to interact with,
    this function returns None to indicate that there should not be a manifest.

    Args:
        app (Application): Pybuild project
        domain (str): Domain that the signals should not be part of
        client_name (str): Client name in the signal database
    Returns:
        spec (dict/None): signal manifest, None if no manifest should be written
    """
    rasters = app.get_rasters()
    LOGGER.debug("Rasters: %s", rasters)
    translation_files = app.get_translation_files()

    dpal = DPAL(app)
    dpal.domain_filter = [domain]
    dpal.parse_definition(translation_files)
    dpal.clear_signal_names()
    dpal.add_signals(app.insignals, "insignals")
    dpal.add_signals(app.outsignals, "outsignals")
    return dpal.to_manifest(client_name)


def configure_parser(parser: argparse.ArgumentParser):
    """Configure parser for generating pt-scheduler wrappers."""
    generation_utils.add_base_args(parser)
    parser.add_argument(
        "--client-name",
        help="Name of the context object in CSP. Defaults to project name."
    )
    parser.add_argument(
        "--dp-interface",
        help="Output file with DP interface specifications",
        type=Path
    )
    parser.add_argument(
        "--dp-manifest-dir",
        help="Output directory for signal manifests",
        type=Path
    )
    parser.add_argument(
        "--hal-interface",
        help="Output file with HAL interface specifications",
        type=Path,
    )
    parser.add_argument(
        "--service-interface",
        help="Output file with service interface specifications",
        type=Path
    )
    parser.set_defaults(func=generate_wrappers_cli)


def generate_wrappers_cli(args: argparse.Namespace):
    """Generate specifications for pt-scheduler wrappers.

    Args:
        args (Namespace): Arguments from command line
    """
    app = generation_utils.process_app(args.config)
    client_name = generation_utils.get_client_name(args, app)
    wrappers(args, app, client_name)


def main(argv: Optional[List[str]] = None):
    """ Main function for stand alone execution.
    Mostly useful for testing and generation of dummy hal specifications
    """
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


def wrappers(args, app, client_name):
    """ Generate specifications for pt-scheduler wrappers.

    Args:
        args (Namespace): Arguments from command line
        app (Application): Application to generate specifications for
        client_name (str): Signal client name
    """
    if args.hal_interface:
        hala = HALA(app)
        interface = generation_utils.get_interface(app, hala)
        interface["relocatable_language"] = "C"
        generation_utils.write_to_file(interface, args.hal_interface, is_yaml=True)
        hals = get_hal_list(app)
        cmake = Path(args.hal_interface.parent, 'hal.cmake')
        generation_utils.write_to_file(hals, cmake)
    if args.dp_interface:
        dp = DPAL(app)
        interface = generation_utils.get_interface(app, dp)
        interface["relocatable_language"] = "C"
        generation_utils.write_to_file(interface, args.dp_interface, is_yaml=True)
    if args.service_interface:
        service = ServiceFramework(app)
        interface = generation_utils.get_interface(app, service)
        interface["relocatable_language"] = "C"
        generation_utils.write_to_file(interface, args.service_interface, is_yaml=True)
        proxies = get_service_list(app)
        cmake = Path(args.service_interface.parent, 'service-proxies.cmake')
        generation_utils.write_to_file(proxies, cmake)
    if args.dp_manifest_dir:
        if not args.dp_manifest_dir.is_dir():
            args.dp_manifest_dir.mkdir()

        # Enable changing client name in generation
        for domain in app.get_domain_names():
            manifest = get_manifest(app, domain, client_name)
            if manifest is None:
                # Manifests without produced or consumed signals are not allowed
                # Therefore get_manifests returned None to tell us to skip this domain
                continue
            domain_file = Path(
                args.dp_manifest_dir,
                f"{domain}_{app.name}.yaml")
            generation_utils.write_to_file(manifest, domain_file, is_yaml=True)


if __name__ == "__main__":
    main(sys.argv[1:])
